from turandot.model import FrontendStrategy, CompanionData, QueueMessage, MessageType
from turandot.ttk.view import TurandotTtkView


class TtkConversionUpdater(FrontendStrategy):
    """Tk specific implementation to update frontend during conversion"""

    STATUS = {
        "idle": "IDLE",
        MessageType.EXCEPTION: "ERROR",
        MessageType.SUCCESS: "SUCCESS",
        MessageType.CANCELED: "CANCELED",
        "conv": "CONVERTING"
    }

    def __init__(self, view: TurandotTtkView):
        self.view = view
        self.warning_string: str = ""

    def _prime_gui(self, msg: QueueMessage):
        """Update GUI at the start of the conversion process"""
        self.view.widgets["open_result_folder"].grid_remove()
        self.view.widgets["status_label_value"].set(self.STATUS["conv"])
        self.view.widgets["n_steps_label_value"].set("0")
        self.view.widgets["total_steps_label_value"].set(str(msg.total_steps))
        self.view.widgets["steps_progress_bar"]["value"] = 0
        self.view.widgets["current_process_description_label_value"].set("")
        self.view.widgets["warnings_title"].grid_remove()
        self.view.widgets["warning_text_view"].grid_remove()
        self.view.widgets["error_title"].grid_remove()
        self.view.widgets["error_type_desc_label"].grid_remove()
        self.view.widgets["error_type_label"].grid_remove()
        self.view.widgets["error_text_view"].grid_remove()
        self.view.widgets["open_result_folder"].grid_remove()
        self.view.widgets["converter_cancel_export_button"].state(["!disabled"])
        self.view.widgets["converter_export_doc_button"].state(["disabled"])

    def _handle_next_step(self, msg: QueueMessage):
        """Update GUI after each step"""
        self.view.widgets["n_steps_label_value"].set(str(msg.n_step))
        self.view.widgets["current_process_description_label_value"].set(msg.step_desc)
        self.view.widgets["steps_progress_bar"]["value"] = ((msg.n_step - 1) / msg.total_steps) * 100

    def _handle_warning(self, msg: QueueMessage):
        """Draw warning onto the GUI"""
        self.warning_string = "{}{}\n".format(self.warning_string, msg.warning)
        self.view.widgets["warning_text_view"].configure(state="normal")
        self.view.widgets["warning_text_view"].delete("1.0", "end")
        self.view.widgets["warning_text_view"].insert('1.0', self.warning_string)
        self.view.widgets["warning_text_view"].configure(state="disabled")
        self.view.widgets["warnings_title"].grid()
        self.view.widgets["warning_text_view"].grid()

    def handle_message(self, msg: QueueMessage):
        if msg.type == MessageType.STARTED:
            self._prime_gui(msg)
        elif msg.type == MessageType.NEXT_STEP:
            self._handle_next_step(msg)
        elif msg.type == MessageType.WARNING:
            self._handle_warning(msg)

    def handle_companion_data(self, data: CompanionData):
        self.view.widgets["status_label_value"].set(self.STATUS[data.status.cause_of_death])
        self.view.widgets["current_process_description_label_value"].set("")
        if data.status.cause_of_death == MessageType.SUCCESS:
            self.view.widgets["steps_progress_bar"]["value"] = 100
            self.view.widgets["open_result_folder"].grid()
        else:
            self.view.widgets["steps_progress_bar"]["value"] = 0
        if data.status.exception is not None:
            self.view.widgets["error_text_view"].configure(state="normal")
            self.view.widgets["error_text_view"].delete("1.0", "end")
            self.view.widgets["error_text_view"].insert(
                '1.0', "{}\n\n{}".format(str(data.status.exception), data.status.exception_tb)
            )
            self.view.widgets["error_text_view"].configure(state="disabled")
            self.view.widgets["error_type_label_value"].set(type(data.status.exception).__name__)
            self.view.widgets["error_title"].grid()
            self.view.widgets["error_type_desc_label"].grid()
            self.view.widgets["error_type_label"].grid()
            self.view.widgets["error_text_view"].grid()
        self.view.widgets["converter_cancel_export_button"].state(["disabled"])
        self.view.widgets["converter_export_doc_button"].state(["!disabled"])



from functools import wraps
from gi.repository import GLib


def guithread(f):
    """Decorator: Push execution to GUI thread"""
    @wraps(f)
    def process(*args, **kwargs):
        GLib.idle_add(f, *args, **kwargs)
    return process



import multiprocessing
from typing import Optional
from pathlib import Path

from turandot import TurandotAssetException
from turandot.model import ConversionAlgorithm, ReferenceSource, JobSettings, ConversionJob, JobAssets, ConversionProcessor, ConverterChain, SourceAsset
from turandot.ui import catch_exception, FrontendUtils, background
from turandot.ttk.controllers import ControllerBase, TtkConversionUpdater
from turandot.ttk.view import TurandotTtkView


class ExportController(ControllerBase):
    """Controller to handle export and cancel export"""

    def __init__(self, view: TurandotTtkView):
        self.msgqueue = multiprocessing.Queue()
        self.processor: Optional[ConversionProcessor] = None
        ControllerBase.__init__(self, view)
        self.conversion_path: Optional[Path] = None
        self._attach_button_events()

    def _attach_button_events(self):
        """Attach callbacks to export/cancel export buttons"""
        self.view.widgets["converter_export_doc_button"].bind("<Button-1>", self._launch_conversion)
        self.view.widgets["converter_cancel_export_button"].bind("<Button-1>", self._cancel_conversion)

    def _collect_job_settings(self) -> JobSettings:
        """Collect settings for conversion job from GUI"""
        convalg: ConversionAlgorithm = self.view.widgets["converter_algorithm_dropdown"].get()
        refsrc: ReferenceSource = self.view.widgets["converter_reference_source_dropdown"].get()
        if refsrc == ReferenceSource.ZOTERO:
            zotlib: Optional[int] = self.view.widgets["converter_zotero_lib_dropdown"].get()
        else:
            zotlib = None
        return JobSettings(convalg, refsrc, zotlib)

    @catch_exception
    def _collect_job(self) -> Optional[ConversionJob]:
        """Collect settings and assets for conversion job from GUI"""
        settings = self._collect_job_settings()
        srcstring = self.view.widgets["converter_source_file_entry_value"].get()
        if srcstring == "":
            raise TurandotAssetException('Field "Source file" must not be empty')
        srcasset = SourceAsset(path=Path(srcstring), expand=True)
        tmplasset = self.view.widgets["converter_template_dropdown"].get()
        if settings.reference_source.value > 1:
            cslasset = self.view.widgets["converter_csl_dropdown"].get()
        else:
            cslasset = None
        if settings.reference_source == ReferenceSource.JSON:
            csljson = self.view.widgets["converter_csljson_entry_value"].get()
            if csljson == "":
                raise TurandotAssetException('Field "CSLJSON file" must not be empty')
        else:
            csljson = None
        assets = JobAssets(srcasset, tmplasset, cslasset, csljson)
        return ConversionJob(assets, settings, self.msgqueue)

    def _create_processor(self) -> Optional[ConversionProcessor]:
        """Create converter object"""
        job = self._collect_job()
        if job is None:
            return None
        chain = ConverterChain.build_chain(job.job_settings.conversion_algorithm)
        frontendstrat = TtkConversionUpdater(self.view)
        return ConversionProcessor(job, chain, frontendstrat)

    @background
    def _fm_open_conversion_folder(self, *args):
        """Use systems file manager to open conversion directory"""
        if self.conversion_path is not None:
            FrontendUtils.fm_open_path(self.conversion_path)

    def _launch_conversion(self, *args):
        """Launch conversion process"""
        self.processor = self._create_processor()
        if self.processor is None:
            return None
        self.conversion_path = self.processor.conversionjob.job_assets.sourcefile.directory
        self.view.widgets["open_result_folder"].bind("<Button-1>", self._fm_open_conversion_folder)
        self.processor.start_conversion()

    def _cancel_conversion(self, *args):
        """Kill conversion process prematurely"""
        self.processor.kill_conversion()



from tkinter import *
from tkhtmlview import HTMLLabel, HTMLScrolledText

from turandot.model import ModelUtils
from turandot.ui import FrontendUtils
from turandot.ttk.controllers import ControllerBase
from turandot.ttk.view import TurandotTtkView, TTKStyles


class AboutController(ControllerBase):
    """Controller to draw text to the 'about' tab"""

    def __init__(self, view: TurandotTtkView):
        super().__init__(view)
        style_obj = TTKStyles()
        self.font_size = style_obj.get_default_font_size()
        txt = self._prepare_markup(ModelUtils.get_asset_content("about.txt"))
        add_to = self.view.widgets["about_container"]
        htlabel = HTMLLabel(add_to, width=60, height=30)
        htlabel.grid(column=0, row=0, sticky=(W, E, N, S))
        htlabel.set_html(txt)

    def _prepare_markup(self, txt: str) -> str:
        """Add markup to text to fix text size"""
        txt = FrontendUtils.replace_version_number(txt)
        lines = txt.splitlines(keepends=False)
        markup_lines = [self._add_size(i) for i in lines if i != ""]
        return "".join(markup_lines)

    def _add_size(self, line: str) -> str:
        """Add p tags and set font sizes"""
        return '<p style="font-size:{}px;">{}</p>'.format(self.font_size, line)



from typing import Union, Optional
from abc import abstractmethod, ABC
from enum import EnumMeta
from tkinter import *
from tkinter import ttk

from turandot.model import ConfigModel
from turandot.ui import background, EnumTranslations
from turandot.ttk.view.styles import TTKStyles
from turandot.ttk.presentations import EnumCombobox


class SettingsViewBase(ABC):
    """Base class for all settings presentations"""

    def __init__(self, label: str, add_to: ttk.Frame, row: int):
        self.label = label
        self.add_to = add_to
        self.row = row

    @abstractmethod
    def _create_label(self) -> ttk.Label:
        """Draw label for settings presentation"""
        pass


class SettingsControlBase(SettingsViewBase, ABC):
    """Base class for writeable settings presentations"""

    def __init__(self, label: str, config_key: Union[str, list], add_to: ttk.Frame, row: int):
        super().__init__(label, add_to, row)
        self.config_key = config_key

    def _create_label(self):
        lbl = ttk.Label(self.add_to, text=self.label)
        lbl.grid(row=self.row, column=0, sticky=W, **TTKStyles.get_padding().left_labels)

    @abstractmethod
    def _callback(self):
        """The presentations callback on a change event"""
        pass


class TitleControl(SettingsViewBase):
    """Read-only presentation to draw a title"""

    def __init__(self, label: str, add_to: ttk.Frame, row: int):
        super().__init__(label=label, add_to=add_to, row=row)
        self._create_label()

    def _create_label(self):
        lbl = ttk.Label(self.add_to, text=self.label, style="h2.TLabel", padding=TTKStyles.get_padding().h2)
        lbl.grid(row=self.row, columnspan=2, sticky=(W, E))


class TextControl(SettingsControlBase):
    """Presentation to write a string to an entry widget"""

    def __init__(self, label: str, config_key: Union[str, list], add_to: ttk.Frame, row: int):
        super().__init__(label, config_key, add_to, row)
        self.entry: Optional[ttk.Entry] = None
        self.sv = StringVar(value=ConfigModel().get_key(self.config_key))
        self._create_label()
        self._create_entry()

    def _create_entry(self):
        """Draw entry widget"""
        self.entry = ttk.Entry(
            self.add_to,
            textvariable=self.sv,
        )
        self.sv.trace_add("write", self._callback)
        self.entry.grid(row=self.row, column=1, sticky=(W, E), **TTKStyles.get_padding().right_entries)

    @background
    def _callback(self, *args):
        ConfigModel().set_key(self.config_key, self.sv.get())


class SpinControl(SettingsControlBase):
    """Presentation to write a number to a spinner widget"""

    def __init__(
        self,
        label: str,
        config_key: Union[str, list],
        add_to: ttk.Frame,
        row: int,
        from_: float = 0x400,
        to: float = 0xFFFF,
        inc: float = 1.0,
        format_: str = "%1.0f",
        wrap: bool = True
    ):
        super().__init__(label, config_key, add_to, row)
        self.spinbox: Optional[ttk.Spinbox] = None
        self.iv = IntVar(value=ConfigModel().get_key(self.config_key))
        self._create_label()
        self._create_spinbox(from_, to, inc, format_, wrap)

    def _unbind_wheel(self):
        """Prevent mouse wheel from doing anything"""
        self.spinbox.unbind_class("TSpinbox", "<ButtonPress-4>")
        self.spinbox.unbind_class("TSpinbox", "<ButtonPress-5>")
        self.spinbox.unbind_class("TSpinbox", "<MouseWheel>")

    def _create_spinbox(self, from_: float, to: float, inc: float, format_: str, wrap: bool):
        """Create spinbox widget"""
        self.spinbox = ttk.Spinbox(
            self.add_to, from_=from_, to=to, increment=inc, format=format_, wrap=wrap, textvariable=self.iv
        )
        self.iv.trace_add("write", self._callback)
        self._unbind_wheel()
        self.spinbox.grid(row=self.row, column=1, sticky=(W, E), **TTKStyles.get_padding().right_entries)

    @background
    def _callback(self, *args):
        ConfigModel().set_key(self.config_key, self.iv.get())


class SwitchControl(SettingsControlBase):
    """Presentation to write a boolean value to a switch"""

    def __init__(self, label: str, config_key: Union[str, list], add_to: ttk.Frame, row: int):
        super().__init__(label, config_key, add_to, row)
        self.bv = BooleanVar(value=ConfigModel().get_key(self.config_key))
        self.checkbutton: Optional[ttk.Checkbutton] = None
        self._create_label()
        self._create_checkbutton()

    def _create_checkbutton(self):
        """Create switch widget"""
        self.checkbutton = ttk.Checkbutton(self.add_to, text="", variable=self.bv)
        self.bv.trace_add("write", self._callback)
        self.checkbutton.grid(row=self.row, column=1, sticky=E, **TTKStyles.get_padding().right_entries)

    @background
    def _callback(self, *args):
        ConfigModel().set_key(self.config_key, self.bv.get())


class TextEnumControl(SettingsControlBase):
    """Create combobox based on the values of en enum"""

    def __init__(self, label: str, config_key: Union[list, str], add_to: ttk.Frame, row: int, model: EnumMeta):
        self.model = model
        super().__init__(label, config_key, add_to, row)
        self.combobox = self._create_combobox(model)
        init_enum_val: model = self.model(ConfigModel().get_key(self.config_key, default=""))
        self.combobox.set(init_enum_val)
        self.combobox.set_callback(self._callback)
        self._create_label()

    def _create_combobox(self, model: EnumMeta) -> EnumCombobox:
        """Draw combobox"""
        combo = EnumCombobox(self.add_to, model, textdict=EnumTranslations.textentries, state="readonly")
        combo.grid(row=self.row, column=1, sticky=(W, E), **TTKStyles.get_padding().right_entries)
        return combo

    @background
    def _callback(self, *args):
        ConfigModel().set_key(self.config_key, self.combobox.get().value)



from turandot.ui.frontend import TurandotFrontend
from turandot.ui.utils import FrontendUtils
from turandot.ui.background import background
from turandot.ui.exceptioncatcher import ExceptionCatcher, CatcherStrategy, catch_exception
from turandot.ui.enumtranslations import EnumTranslations



from abc import ABC, abstractmethod
from turandot.gtk import TurandotGtkView


class BaseController(ABC):
    """Base class for GTK controllers"""

    def __init__(self, view: TurandotGtkView):
        self.view = view

    @abstractmethod
    def connect_events(self):
        """Connect event handlers to GUI"""
        pass



class TurandotException(Exception):
    """Exception base class for Turandot exceptions"""
    pass


class TurandotAssetException(TurandotException):
    """Exception thrown on asset expansion"""
    pass


class TurandotConversionException(TurandotException):
    """Exception thrown on errors during conversion"""
    pass


class TurandotConnectionException(TurandotException):
    """Exception thrown if an API connection fails"""
    pass


class TurandotCiteprocException(TurandotException):
    """Exception thrown on errors by citeproc conversion"""
    pass



import gi
gi.require_version('Gtk', '3.0')
from typing import Union
from tempfile import NamedTemporaryFile
from gi.repository import Gtk
from turandot.model import ModelUtils
from turandot.gtk.presentations import DatabaseDropdown, EnumDropdown
from turandot.model import TemplateAsset, CslAsset, ReferenceSource, ConversionAlgorithm


class TurandotGtkView:
    """Launch GTK application"""

    @staticmethod
    def _get_glade() -> str:
        """Get Glade file from assets"""
        return ModelUtils.get_asset_content("mainwindow.glade")

    @staticmethod
    def _get_icon() -> NamedTemporaryFile:
        """Dump icon to a temp file to use as GTK icon"""
        icon = NamedTemporaryFile(mode='w')
        icon.write(ModelUtils.get_asset_content("turandot.svg"))
        return icon

    def __init__(self):
        glade_str = self._get_glade()
        builder = Gtk.Builder()
        self.elements = builder.new_from_string(glade_str, len(glade_str))
        self.datadropdowns: dict[str, DatabaseDropdown] = {
            "template_dropdown": DatabaseDropdown(model=TemplateAsset),
            "template_editor_dropdown": DatabaseDropdown(model=TemplateAsset, new_option=True),
            "csl_dropdown": DatabaseDropdown(model=CslAsset),
            "csl_editor_dropdown": DatabaseDropdown(model=CslAsset, new_option=True)
        }
        self.enumdropdowns: dict[str, EnumDropdown] = {
            "conversion_algorithm_dropdown": EnumDropdown(ConversionAlgorithm),
            "reference_source_dropdown": EnumDropdown(ReferenceSource)
        }
        self.addressable_fields: dict[str, Gtk.Widget] = {}
        self.window: Gtk.ApplicationWindow = self.elements.get_object("mainwindow")
        self._add_datadropdowns()
        self._add_enumdropdowns()

    def _add_datadropdowns(self):
        """Add database-fed dropdowns to GUI"""
        self.get_element("template_dropdown_container").add(self.get_element("template_dropdown"))
        self.get_element("csl_dropdown_container").add(self.get_element("csl_dropdown"))
        self.get_element("template_editor_dropdown_container").add(self.get_element("template_editor_dropdown"))
        self.get_element("csl_editor_dropdown_container").add(self.get_element("csl_editor_dropdown"))
        self.get_element("template_dropdown").show()
        self.get_element("template_editor_dropdown").show()
        self.get_element("csl_editor_dropdown").show()

    def _add_enumdropdowns(self):
        """Add enum-based dropdowns to the GUI"""
        self.get_element("conversion_algorithm_dropdown_container").add(self.get_element("conversion_algorithm_dropdown"))
        self.get_element("conversion_algorithm_dropdown").show()
        self.get_element("reference_source_dropdown_container").add(self.get_element("reference_source_dropdown"))
        self.get_element("reference_source_dropdown").show()

    def get_element(self, element_id: str) -> Union[DatabaseDropdown, EnumDropdown, Gtk.Widget, None]:
        """Get widget from view"""
        dd = self.datadropdowns.get(element_id, None)
        if dd is not None:
            return dd
        ed = self.enumdropdowns.get(element_id, None)
        if ed is not None:
            return ed
        af = self.addressable_fields.get(element_id, None)
        if af is not None:
            return af
        return self.elements.get_object(element_id)

    def show_main(self):
        """Show main window"""
        self.window.connect("destroy", Gtk.main_quit)
        icon = self._get_icon()
        self.window.set_icon_from_file(icon.name)
        icon.close()
        self.window.show()
        Gtk.main()



import os

from turandot import SysInfo, OpSys
from turandot.ui import TurandotFrontend
from turandot.model import ModelUtils


class TurandotMain:

    def __init__(self):
        import multiprocessing
        # Solves multiple app start bug when frozen
        multiprocessing.freeze_support()
        # Drastically increases stability of detached Weasyprint conversion process (but lengthens process init time)
        multiprocessing.set_start_method("spawn")
        # Get OS infos
        self.sysinfo = SysInfo()
        # Windows-specific stuff
        if self.sysinfo.system == OpSys.WINDOWS:
            self._prep_windows()
        if self.sysinfo.system == OpSys.LINUX:
            try:
                from turandot.gtk import TurandotGtk
                self.frontend: TurandotFrontend = TurandotGtk()
                return
            except ImportError:
                pass
        from turandot.ttk import TurandotTtk
        self.frontend: TurandotFrontend = TurandotTtk()

    @staticmethod
    def _prep_windows():
        """Create fontconfig file on Windows, not really sure if this does any good"""
        target = ModelUtils.get_config_dir() / "fonts.conf"
        if not target.is_file():
            with target.open('w') as fh:
                fh.write(ModelUtils.get_asset_content("windows_fonts.conf"))
        """Set Windows OS specific env variables, maybe obsolete"""
        os.environ['PYTHONUTF8'] = "1"
        os.environ['FONTCONFIG_FILE'] = str(ModelUtils.get_config_dir() / "fonts.conf")

    def main(self):
        self.frontend.run()


def launch():
    """Function to call from pynsist"""
    TurandotMain().main()


if __name__ == "__main__":
    launch()



import threading
from functools import wraps
from turandot import Singleton


def background(f):
    """Decorator: Push execution to background thread"""
    @wraps(f)
    def process(*args, **kwargs):
        processor = BackgroundProcessor()
        new_thread = threading.Thread(target=processor.run_task_blocking, args=[f, *args], kwargs=kwargs)
        new_thread.start()
    return process


class BackgroundProcessor(metaclass=Singleton):
    """Singleton executor to run threads in the background"""

    def __init__(self):
        self.semaphore = threading.Semaphore()

    def run_task_blocking(self, f, *args, **kwargs):
        """Run background tasks one after another"""
        self.semaphore.acquire()
        f(*args, **kwargs)
        self.semaphore.release()



from turandot.model import ModelUtils
from turandot.ui import FrontendUtils
from turandot.gtk.controllers import BaseController


class AboutController(BaseController):
    """Controller for the 'About' tab, mainly draws text to GUI"""

    def connect_events(self):
        self._draw_text()

    def _draw_text(self):
        """Get text from assets and draw to 'About' tab"""
        txt = ModelUtils.get_asset_content("about.txt")
        txt = FrontendUtils.replace_version_number(txt)
        self.view.get_element("about_label").set_markup(txt)



from enum import Enum
from md_citeproc import NotationStyle, OutputStyle

from turandot.model import ReferenceSource, ConversionAlgorithm


class EnumTranslations:
    """Human readable text for each entry of an enum dropdown"""

    textentries: dict[Enum, str] = {
        ReferenceSource.NOTHING: "No automatic processing",
        ReferenceSource.NOSOURCE: "Processing without data source",
        ReferenceSource.ZOTERO: "Zotero",
        ReferenceSource.JSON: "CSLJSON file",

        ConversionAlgorithm.WEASYPRINT: "Weasyprint/PDF",

        NotationStyle.INLINE: "inline",
        NotationStyle.FOOTNOTE: "footnotes",
        NotationStyle.INLINE_FOOTNOTE: "inline-footnotes",

        OutputStyle.INLINE: "inline",
        OutputStyle.NUM_FOOTNOTES: "enumerated footnotes"
    }

    inversion = dict((v, k) for k, v in textentries.items())

    @staticmethod
    def get(key: Enum):
        """Get human readable text"""
        return EnumTranslations.textentries.get(key, key.name)

    @staticmethod
    def inverse_get(value: str):
        """
        This is a really dumb method to do this
        but ttk's combobox widgets are kind of forcing this method
        """
        return EnumTranslations.inversion.get(value, None)



from typing import Optional, Union
from pathlib import Path
from turandot.model import TemplateAsset, CslAsset
from turandot.ui import catch_exception
from turandot.gtk import DropdownPreSelections, TurandotGtkView, TypeFilterEnum
from turandot.gtk.controllers import BaseController, DialogController
from turandot.gtk.presentations import DatabaseDropdown


class DataDropdownController(BaseController):
    """Controller for 'Template' and 'CSL' tabs"""

    def __init__(self, view: TurandotGtkView):
        BaseController.__init__(self, view=view)
        self._update_templates(preselect=DropdownPreSelections.FIRST_ELEMENT)
        self._update_csl(preselect=DropdownPreSelections.FIRST_ELEMENT)

    def _update_templates(self, preselect: Union[DropdownPreSelections, int, type(None)] = None):
        """Update entries in template dropdowns"""
        self.view.get_element("template_dropdown").update(preselect=DropdownPreSelections.SAME_ELEMENT)
        self.view.get_element("template_editor_dropdown").update(preselect=preselect)

    def _update_csl(self, preselect: Union[DropdownPreSelections, int, type(None)] = None):
        """Update entries in CSL dropdowns"""
        self.view.get_element("csl_dropdown").update(preselect=DropdownPreSelections.SAME_ELEMENT)
        self.view.get_element("csl_editor_dropdown").update(preselect=preselect)

    def connect_events(self):
        self.view.get_element("template_editor_dropdown").set_change_callback(self._on_template_editor_dropdown_change)
        self.view.get_element("csl_editor_dropdown").set_change_callback(self._on_csl_editor_dropdown_change)
        self.view.get_element("tmpl_base_select_button").connect("clicked", self._browse_template)
        self.view.get_element("csl_select_button").connect("clicked", self._browse_csl)
        self.view.get_element("tmpl_save_button").connect("clicked", self._save_template)
        self.view.get_element("tmpl_delete_button").connect("clicked", self._delete_template)
        self.view.get_element("csl_save_button").connect("clicked", self._save_csl)
        self.view.get_element("csl_delete_button").connect("clicked", self._delete_csl)

    def _on_template_editor_dropdown_change(self, *args):
        """Draw template settings to 'Template' tab"""
        asset = self.view.get_element("template_editor_dropdown").get_selected_asset()
        if asset is None:
            self.view.get_element("tmpl_delete_button").set_sensitive(False)
            self.view.get_element("tmpl_base_file_enty").set_text("")
            self.view.get_element("tmpl_allow_jinja").set_active(False)
            self.view.get_element("tmpl_allow_mako").set_active(False)
        else:
            self.view.get_element("tmpl_delete_button").set_sensitive(True)
            self.view.get_element("tmpl_base_file_enty").set_text(str(asset.path))
            self.view.get_element("tmpl_allow_jinja").set_active(asset.allow_jinja)
            self.view.get_element("tmpl_allow_mako").set_active(asset.allow_mako)

    def _on_csl_editor_dropdown_change(self, *args):
        """Draw CSL settings to 'CSL' tab"""
        asset = self.view.get_element("csl_editor_dropdown").get_selected_asset()
        if asset is None:
            self.view.get_element("csl_delete_button").set_sensitive(False)
            self.view.get_element("csl_base_file_enty").set_text("")
        else:
            self.view.get_element("csl_delete_button").set_sensitive(True)
            self.view.get_element("csl_base_file_enty").set_text(str(asset.path))

    def _browse_template(self, *args):
        """Create file chooser dialog to pick template base file"""
        DialogController.db_fed_file_chooser_dialog(
            view=self.view,
            return_entry_id="tmpl_base_file_enty",
            typefilter=TypeFilterEnum.TMPL,
            title="Select template base"
        )

    def _browse_csl(self, *args):
        """Create file chooser dialog to pick csl base file"""
        DialogController.db_fed_file_chooser_dialog(
            view=self.view,
            return_entry_id="csl_base_file_enty",
            typefilter=TypeFilterEnum.CSL,
            title="Select csl file"
        )

    @catch_exception
    def _save_template(self, *args):
        """Save data from GUI as template, update menu"""
        asset = self.view.get_element("template_editor_dropdown").get_selected_asset()
        p = Path(self.view.get_element("tmpl_base_file_enty").get_text())
        aj = self.view.get_element("tmpl_allow_jinja").get_active()
        am = self.view.get_element("tmpl_allow_mako").get_active()
        if asset is None:
            asset = TemplateAsset(p, am, aj)
        else:
            asset.path = p
            asset.allow_mako = am
            asset.allow_jinja = aj
        asset.save()
        self._update_templates(preselect=asset.dbid)

    def _delete_template(self, *args):
        """Show delete confirmation dialog"""
        DialogController.show_confirm_dialog(
            self.view,
            "Remove template entry?\n(No files will be deleted)",
            self._delete_template_confirmed
        )

    def _delete_template_confirmed(self, *args):
        """Delete selected template from database, update menu"""
        self.view.get_element("confirm_dialog").hide()
        asset = self.view.get_element("template_editor_dropdown").get_selected_asset()
        if asset is not None:
            asset.delete()
        self._update_templates(DropdownPreSelections.FIRST_ELEMENT)

    @catch_exception
    def _save_csl(self, *args):
        """Save data from GUI as CSL entry, update menus"""
        asset = self.view.get_element("csl_editor_dropdown").get_selected_asset()
        p = Path(self.view.get_element("csl_base_file_enty").get_text())
        if asset is None:
            asset = CslAsset(p)
        else:
            asset.path = p
        asset.save()
        self._update_csl(preselect=asset.dbid)

    def _delete_csl(self, *args):
        """Show delete confirmation dialog"""
        DialogController.show_confirm_dialog(
            self.view,
            "Remove csl entry?\n(No files will be deleted)",
            self._delete_csl_confirmed
        )

    def _delete_csl_confirmed(self, *args):
        """Delete selected csl entry from database, update menus"""
        self.view.get_element("confirm_dialog").hide()
        asset = self.view.get_element("csl_editor_dropdown").get_selected_asset()
        if asset is not None:
            asset.delete()
        self._update_csl(DropdownPreSelections.FIRST_ELEMENT)



from tkinter import *
from tkinter import ttk

from turandot.ttk.view import ViewComponent


class AboutTab(ViewComponent):
    """View component for 'About' tab"""

    def _create_about_frame(self, aframe: ttk.Frame):
        self.widgets["about_container"] = ttk.Frame(aframe)
        self.widgets["about_container"].grid(row=0, column=0, sticky=(W, E, S, N))

    def create(self, add_to: ttk.Frame) -> ttk.Frame:
        aframe = ttk.Frame(add_to)
        aframe.grid(column=0, row=0, sticky=(N, E, S, W))
        self._create_about_frame(aframe)
        return aframe



import requests
import json
from enum import Enum
from turandot import TurandotConnectionException
from turandot.model import ConfigDict


class ZoteroDataFormat(Enum):
    BIBTEX = "biblatex"
    CSLJSON = "csljson"


class ZoteroConnector:
    """Connect to Zoteros BetterBibtex extension to get bibliographic data"""

    def __init__(self, config: ConfigDict):
        self.hostname = config.get_key(['api', 'zotero', 'host'], "localhost")
        self.port = config.get_key(['api', 'zotero', 'port'], 23119)
        self.connection_error = TurandotConnectionException(
            "No answer from {}:{}. Zotero and/or Better Bibtex extension not running?".format(self.hostname, self.port)
        )

    def _assemble_url(self, endpoint):
        """Assemble endpoint specific url"""
        return "http://{}:{}{}".format(self.hostname, self.port, endpoint)

    def get_libraries(self) -> list:
        """
        Get list of zotero libs with their IDs
        ct. https://github.com/retorquere/zotero-better-bibtex/issues/787
        :return: List of Zotero Libraries
        """
        endpoint = "/better-bibtex/json-rpc"
        payload = json.dumps({"jsonrpc": "2.0", "method": "user.groups"})
        headers = {'content-type': 'application/json'}
        url = self._assemble_url(endpoint)
        try:
            response = requests.post(url, data=payload, headers=headers)
            libs = json.loads(response.content.decode())
            return libs["result"]
        except requests.exceptions.ConnectionError:
            raise self.connection_error

    def get_bib_data(self, libid: int, form: ZoteroDataFormat) -> str:
        """
        Get bibliographic data from Zotero
        :param libid: Library to get data from
        :param form: Format: JSON or Bibtex
        :return: Bibliographic data in text format
        """
        endpoint = "/better-bibtex/library?/{}/library.{}".format(libid, form.value)
        url = self._assemble_url(endpoint)
        try:
            response = requests.get(url)
            return response.content.decode()
        except requests.exceptions.ConnectionError:
            raise self.connection_error

    def get_csljson(self, libid: int) -> str:
        """Get CSLJSON string from Zotero"""
        return self.get_bib_data(libid=libid, form=ZoteroDataFormat.CSLJSON)



from enum import Enum, auto


class OpSys(Enum):
    """Enum representing the running OS"""
    LINUX = 0
    WINDOWS = 1
    MACOS = 2


class Architecture(Enum):
    """Enum representing the running CPU architecture"""
    AMD64 = 0


class ReferenceSource(Enum):
    """Enum representing the source of bibliographic data to be used"""
    NOTHING = 0
    NOSOURCE = 1
    ZOTERO = 2
    JSON = 3


class TemplatingEngine(Enum):
    """Enum representing the templating engine to be used for the conversion"""
    NOTHING = 0
    JINJA2 = 1
    MAKO = 2


class ConversionAlgorithm(Enum):
    """Enum representing the conversion algorithm to be used for the conversion"""
    WEASYPRINT = 0


class OptionalStage(Enum):
    """Enum representing the stage to attach a optional converter to a mandatory one"""
    PRE = 0
    POST = 1


class MessageType(Enum):
    """Enum representing the type of a message in the queue from backend to frontend"""
    STARTED = auto()
    COPY = auto()
    EXCEPTION = auto()
    WARNING = auto()
    NEXT_STEP = auto()
    SUCCESS = auto()
    DIED = auto()
    CANCELED = auto()



from abc import ABC, abstractmethod
from turandot.ttk.view import TurandotTtkView


class ControllerBase(ABC):
    """Abstract class to subclass controllers from"""

    def __init__(self, view: TurandotTtkView):
        self.view = view



import traceback
import threading
import multiprocessing
from typing import Optional
from turandot.model import ConversionJob, ConverterBase, CompanionData, FrontendStrategy, QueueMessage, MessageType


class ConversionProcessor:
    """
    Processor class to work through conversion chain.
    Conversion is detached in a separate process to make it easily killable.
    Spawn a watchdog thread to wait for subprocess to finish.
    Spawn queue worker thread to read messages from the queue to update frontend
    """

    def __init__(self, conversionjob: ConversionJob, converterchain: ConverterBase, frontendstrat: FrontendStrategy):
        self.conversionjob: ConversionJob = conversionjob
        self.msgqueue: multiprocessing.Queue = self.conversionjob.msgqueue
        self.converterchain: ConverterBase = converterchain
        self.frontendstrat: FrontendStrategy = frontendstrat
        self.conversionproc: Optional[multiprocessing.Process] = None
        self.companiondata: CompanionData = CompanionData(self.msgqueue)

    def _watchdog_worker(self):
        """Check if conversion process still alive, notify queue on death"""
        self.conversionproc.join()
        self.msgqueue.put(QueueMessage.deathmsg())

    def _queue_worker(self):
        """Get messages from queue until conversion dies"""
        while 1:
            msg = self.msgqueue.get(block=True)
            self._process_queue_msg(msg)
            if msg.type == MessageType.DIED:
                break
        self._cleanup()
        self.frontendstrat.handle_companion_data(self.companiondata)

    def _process_queue_msg(self, msg: QueueMessage):
        """Pass queue messages to companion data log and frontend"""
        # Immediately update frontend
        if msg.type in [MessageType.STARTED, MessageType.NEXT_STEP, MessageType.WARNING, MessageType.DIED]:
            if msg.type == MessageType.WARNING:
                self.companiondata.status.warnings.append(msg.warning)
            self.frontendstrat.handle_message(msg)
        # Save data, do something later
        if msg.type == MessageType.COPY:
            self.companiondata.copylog.append(msg.copy)
        if msg.type == MessageType.EXCEPTION:
            self.companiondata.status.exception = msg.exception
            self.companiondata.status.exception_tb = msg.traceback
            self.companiondata.status.cause_of_death = msg.type
        if msg.type == MessageType.SUCCESS:
            self.companiondata.status.cause_of_death = msg.type

    def _cleanup(self):
        """Cleanup all copied files on death"""
        self.companiondata.copylog.delete_all()

    @staticmethod
    def _detached_conversion(conversionjob: ConversionJob, converterchain: ConverterBase):
        """Process complete conversion in different process"""
        try:
            converterchain.handle(conversionjob)
            conversionjob.msgqueue.put(QueueMessage.successmsg())
        except Exception as e:
            tb = traceback.format_exc()
            print(tb)
            conversionjob.msgqueue.put(QueueMessage.exceptionmsg(e, tb))

    def start_conversion(self):
        """Branch off in threads and processes, start conversion"""
        self.msgqueue.put(QueueMessage.startedmsg(self.converterchain.total_steps))
        self.conversionproc = multiprocessing.Process(
            target=self._detached_conversion,
            args=(self.conversionjob, self.converterchain)
        )
        self.conversionproc.start()
        threading.Thread(target=self._queue_worker).start()
        threading.Thread(target=self._watchdog_worker).start()

    def kill_conversion(self):
        """Kill process by force"""
        if self.conversionproc is not None and self.conversionproc.is_alive():
            self.conversionproc.kill()



from turandot.model.datastructures.baseasset import BaseAsset


class TextAsset(BaseAsset):
    """Base class to represent a conversion asset consisting of a single text file"""

    def expand(self):
        self._read_path()
        self._read_content()



from turandot import Singleton
from pathlib import Path
from copy import deepcopy
from typing import Union, Any
import ruamel.yaml
from turandot.model import ConfigDict, ModelUtils


class ConfigModel(metaclass=Singleton):
    """Interact with Config file: Write and read config values"""

    @staticmethod
    def _initialize_default(target: Path):
        """If no config file present, init the default"""
        ModelUtils.get_config_dir().mkdir(exist_ok=True)
        with target.open('w') as fh:
            fh.write(ModelUtils.get_asset_content("default.yaml"))

    def __init__(self):
        self.yaml = ruamel.yaml.YAML(typ="rt", pure=True)
        self.path = ModelUtils.get_config_file()
        if not self.path.is_file():
            ConfigModel._initialize_default(self.path)
        self.data = self.yaml.load(self.path)

    def get_key(self, key: Union[list, str], default=None):
        """
        Get config key from file
        :param key: Key string or list of key for a nested value
        :param default: default to return if key is not found
        :return: config value
        """
        if type(key) != list:
            return self.data.get(key, default)
        else:
            search_space = self.data
            for i in key:
                if i in search_space.keys():
                    search_space = search_space.get(i)
                else:
                    return default
            return search_space

    def set_key(self, key: Union[list, str], val: Any):
        """
        Set config key in config file
        :param key: Key string or list of key for a nested value
        :param val: Value to set
        :return: None
        """
        if type(key) == str:
            self.data[key] = val
        else:
            search_space = self.data
            for i in key[:-1]:
                search_space = search_space.get(i)
            search_space[key[-1]] = val
        self.yaml.dump(self.data, self.path)

    def get_dict(self) -> ConfigDict:
        """
        Get a copy of the complete config dict (editing the copy does not edit the config!)
        :return: Deep copy of the complete config
        """
        return ConfigDict.from_ordered(deepcopy(self.data))



from abc import ABC, abstractmethod
from turandot.model import QueueMessage, CompanionData


class FrontendStrategy(ABC):
    """
    Interface to be implemented by the frontend
    Allows state notification from the backend
    """

    @abstractmethod
    def handle_message(self, msg: QueueMessage):
        """Handle a Queue message from the backend: Do something with it on the frontend"""
        pass

    @abstractmethod
    def handle_companion_data(self, data: CompanionData):
        """Handle a companion data object after conversion: Do something with it on the frontend"""
        pass


class PrintStrategy(FrontendStrategy):
    """Dummy implementation of frontend strategy"""

    def handle_message(self, msg: QueueMessage):
        print(msg)

    def handle_companion_data(self, data: CompanionData):
        print(data)



from copy import deepcopy
import markdown
from importlib import import_module
from turandot.model import ConverterBase, ConversionJob
from markdown_katex.extension import KatexExtension
from md_citeproc import CiteprocExtension
from typing import Optional

from turandot.model import ReferenceSource, QueueMessage


class ConvertToHtml(ConverterBase):
    """Convert preprocessed markdown file to incomplete html body without proper containment"""

    NON_BUILTIN = ['markdown_katex', 'md_citeproc']
    DESC: str = "Converting Markdown to HTML"
    PREFIX = "posthtml_"
    POSTFIX = ".html"
    SAVABLE: bool = True

    def __init__(self):
        self.extensions: list = []
        self.config_builtin: dict = {}
        self.config_non_builtin: dict = {}
        self.citeproc_ext: Optional[CiteprocExtension] = None
        ConverterBase.__init__(self)

    def _load_config(self):
        """Load markdown extension config"""
        ext_config = deepcopy(self.conversion_job.config.get_key(["processors", "convert_to_html", "markdown_ext"], None))
        if ext_config is not None:
            for k, v in ext_config.items():
                if k in ConvertToHtml.NON_BUILTIN:
                    self.config_non_builtin.update({k: v})
                else:
                    self.config_builtin.update({k: v})

    def _create_builtin_extensions(self):
        """Create built-in markdown extensions"""
        for k, v in self.config_builtin.items():
            if v.pop("enable", False):
                mod = import_module("markdown.extensions.{}".format(k), package="markdown")
                ext = mod.makeExtension(**v)
                self.extensions.append(ext)

    def _create_special_extensions(self):
        """Create third party markdown extensions"""
        for k, v in self.config_non_builtin.items():
            # Construct Katex extension
            if k == "markdown_katex" and v.pop("enable", False):
                ext = KatexExtension(**v)
                self.extensions.append(ext)
            # Construct Citeproc extension
            if k == "md_citeproc" and v.pop("enable", False) \
                    and (self.conversion_job.job_settings.reference_source != ReferenceSource.NOTHING):
                self.citeproc_ext = self._create_citeproc_extension(v)
                self.extensions.append(self.citeproc_ext)

    def _create_citeproc_extension(self, citeproc_config) -> CiteprocExtension:
        """
        Explicitly construct CiteprocExtension
        Override csljson, cslfile with values from conversion job
        """
        if self.conversion_job.job_settings.reference_source == ReferenceSource.NOSOURCE:
            citeproc_config["csljson"] = None
            citeproc_config["cslfile"] = None
        else:
            citeproc_config["csljson"] = self.conversion_job.job_assets.csljson.path
            citeproc_config["cslfile"] = self.conversion_job.job_assets.cslfile.path
        return CiteprocExtension(**citeproc_config)

    def process_step(self) -> ConversionJob:
        self._load_config()
        self._create_builtin_extensions()
        self._create_special_extensions()
        self.conversion_job.current_step.content = markdown.markdown(
            self.conversion_job.current_step.content,
            extensions=self.extensions
        )
        if self.citeproc_ext is not None:
            warnings = self.citeproc_ext.get_warnings()
            for i in warnings:
                self.conversion_job.msgqueue.put(QueueMessage.warningmsg(str(i)))
        return self.conversion_job



from typing import Optional, Callable
from enum import EnumMeta
import tkinter as tk
from tkinter import ttk

from turandot.ttk.presentations import KvCombobox


class EnumCombobox(KvCombobox):
    """Key-value combobox to return enum values on selection"""

    def __init__(self, master: tk.BaseWidget, model: EnumMeta, textdict: Optional[dict] = None, **kwargs):
        super().__init__(master, **kwargs)
        for i in model:
            if textdict is None:
                self.add_option(i, str(i.value))
            else:
                self.add_option(i, textdict.get(i, i.value))



import os
from tempfile import NamedTemporaryFile
from tkinter import *
from tkinter import ttk

from turandot.model import ModelUtils
from turandot.ttk.view import TTKStyles
from turandot.ttk.view import ViewComponent
from turandot.ttk.view.convertertab import ConverterTab
from turandot.ttk.view.templatetab import TemplateTab
from turandot.ttk.view.csltab import CslTab
from turandot.ttk.view.settingstab import SettingsTab
from turandot.ttk.view.abouttab import AboutTab
from turandot.ttk.view.processframe import ProcessFrame


class TurandotTtkView:
    """Create view for TK GUI"""

    def __init__(self):
        self.widgets = {}
        self.root = Tk()
        self.icon = TurandotTtkView._get_icon()
        pi = PhotoImage(file=self.icon.name)
        self.root.iconphoto(False, pi)
        self.icon.close()
        os.unlink(self.icon.name)
        self.root.title("Turandot")
        self.root.columnconfigure(0, weight=1)
        self.root.columnconfigure(0, weight=1)
        TTKStyles().add_styles(self.root)
        self.mainframe = ttk.Frame(self.root, padding="0 5 0 0")
        self.mainframe.grid(column=0, row=0, sticky=(N, W, E, S))
        self.mainframe.columnconfigure(0, weight=3)
        self.mainframe.columnconfigure(1, weight=2)
        self.mainframe.rowconfigure(0, weight=1)
        self._assemble_gui()

    @staticmethod
    def _get_icon() -> NamedTemporaryFile:
        """Dump icon to a temp file to use as TK icon"""
        icon = NamedTemporaryFile(mode='wb', delete=False)
        icon.write(ModelUtils.get_asset_bytes("turandot.png"))
        return icon

    def _add_component(self, component: ViewComponent, add_to: Frame) -> ttk.Frame:
        """Add the widgets of a view component to the view"""
        f = component.create(add_to=add_to)
        self.widgets = self.widgets | component.widgets
        return f

    def _create_notebook(self):
        """Create basic notebook widgets to add view components to"""
        notebook = ttk.Notebook(self.mainframe, width=500)
        notebook.grid(column=0, row=0, sticky=(N, W, E, S))
        notebook.columnconfigure(0, weight=1)
        notebook.rowconfigure(0, weight=1)
        # Create Converter Tab
        converter_frame = ttk.Frame(notebook, padding=TTKStyles.get_padding().notebook_frames)
        converter_frame.grid(column=0, row=0, sticky=(N, W, E, S))
        converter_frame.columnconfigure(0, weight=1)
        converter_frame.rowconfigure(0, weight=1)
        # Create Frame in Converter Tab
        self._add_component(ConverterTab(), converter_frame)
        notebook.add(converter_frame, text="Converter")
        # Create Template Tab
        template_frame = ttk.Frame(notebook, padding=TTKStyles.get_padding().no_title_notebook_frames)
        template_frame.grid(column=0, row=0, sticky=(N, E, W, S))
        template_frame.columnconfigure(0, weight=1)
        # Create Frame in Template Tab
        self._add_component(TemplateTab(), template_frame)
        notebook.add(template_frame, text="Templates")
        # Create CSL Tab
        csl_frame = ttk.Frame(notebook, padding=TTKStyles.get_padding().no_title_notebook_frames)
        csl_frame.grid(column=0, row=0, sticky=(N, E, S, W))
        csl_frame.columnconfigure(0, weight=1)
        # Create Frame in CSL Tab
        self._add_component(CslTab(), csl_frame)
        notebook.add(csl_frame, text="CSL")
        # Create Settings Tab
        settings_frame = ttk.Frame(notebook, padding=TTKStyles.get_padding().notebook_frames)
        settings_frame.grid(column=0, row=0, sticky=(N, E, S, W))
        settings_frame.columnconfigure(0, weight=1)
        # Create Frame in Settings Tab
        self._add_component(SettingsTab(), settings_frame)
        notebook.add(settings_frame, text="Settings")
        # Create About Tab
        about_frame = ttk.Frame(notebook, padding=TTKStyles.get_padding().about_notebook_frame)
        about_frame.grid(column=0, row=0, sticky=(N, E, S, W))
        about_frame.columnconfigure(0, weight=1)
        # Create Frame in Settings tab
        self._add_component(AboutTab(), about_frame)
        notebook.add(about_frame, text="About")

    def _create_processframe(self):
        """Add process frame widgets from process frame component"""
        pframe = ttk.Frame(self.mainframe, padding="4 0 12 8")
        pframe.grid(row=0, column=1, sticky=(N, E, S, W))
        pframe.columnconfigure(0, weight=1)
        pframe.rowconfigure(0, weight=1)
        self._add_component(ProcessFrame(), pframe)

    def _assemble_gui(self):
        """Assemble complete view from components"""
        self._create_notebook()
        self._create_processframe()

    def run(self):
        """Run tk main loop after view construction"""
        self.root.mainloop()



import os
import shutil
from pathlib import Path
import zipfile
from turandot import TurandotConversionException
from turandot.model import ConverterBase, ConversionJob, QueueMessage


class CopyTemplate(ConverterBase):
    """Copy all files from the template to conversion directory, log for deletion afterwards"""

    DESC: str = "Copying template data"

    @staticmethod
    def _make_abs_path(relpath: str, destpath: Path) -> Path:
        """Complete relative target path by appending it to conversion directory"""
        for i in relpath.split("/"):
            if i is not None:
                destpath = destpath / i
        return Path(destpath)

    # Recursive Copy & Log
    def _copytree(self, src: Path, dst: Path):
        """Recursively copy a directory tree and log copied files"""
        if not dst.is_dir():
            dst.mkdir()
            self.conversion_job.msgqueue.put(QueueMessage.copymsg((dst, 'd')))
        for item in src.iterdir():
            d = dst / item.name
            if item.is_dir():
                self._copytree(item, d)
            elif item.is_file():
                if d.is_file():
                    raise TurandotConversionException("Template file copy conflict on '{}'".format(d))
                shutil.copy(item, d)
                self.conversion_job.msgqueue.put(QueueMessage.copymsg((d, 'f')))

    def _extract(self, sourcefile: Path, destination: Path):
        """Extract a zip archive and log copied contents"""
        zf = zipfile.ZipFile(sourcefile)
        for i in zf.namelist():
            absp = self._make_abs_path(i, destination)
            if i[-1] == "/":
                if not absp.is_dir():
                    absp.mkdir()
                    self.conversion_job.msgqueue.put(QueueMessage.copymsg((absp, 'd')))
            else:
                if absp.is_file():
                    raise TurandotConversionException("Template file copy conflict on '{}'".format(absp))
                else:
                    with zipfile.ZipFile(sourcefile) as z:
                        z.extract(i, str(destination))
                        self.conversion_job.msgqueue.put(QueueMessage.copymsg((absp, 'f')))

    def process_step(self) -> ConversionJob:
        if self.conversion_job.job_assets.template.zipped:
            self._extract(
                sourcefile=self.conversion_job.job_assets.template.path,
                destination=self.conversion_job.job_assets.sourcefile.directory
            )
        else:
            self._copytree(
                src=self.conversion_job.job_assets.template.directory,
                dst=self.conversion_job.job_assets.sourcefile.directory
            )
        return self.conversion_job



from peewee import CharField, BooleanField, TextField, PrimaryKeyField, AutoField
from turandot.model.db.dbinit import BaseModel


class Csl(BaseModel):
    """Save path of CSL files to database"""
    dbid = AutoField(primary_key=True)
    path = TextField()


class FileSelectPersistence(BaseModel):
    """Save recently used path of file select dialogs to database"""
    input_id = TextField(primary_key=True)
    last_path = TextField(null=True)


class Templates(BaseModel):
    """Save path & allowed templating engines of templates to the database"""
    dbid = AutoField(primary_key=True)
    path = TextField()
    allow_jinja = BooleanField()
    allow_mako = BooleanField()


# Make sure tables are created before referenced
t = Templates()
c = Csl()
f = FileSelectPersistence()



import multiprocessing
from typing import Optional
from pathlib import Path
from turandot import TurandotAssetException
from turandot.model import ConversionAlgorithm, ReferenceSource, JobSettings, ConversionJob, JobAssets, ConversionProcessor, ConverterChain, SourceAsset
from turandot.ui import catch_exception, FrontendUtils
from turandot.gtk import TurandotGtkView
from turandot.gtk.controllers import BaseController, GtkConversionUpdater


class ExportController(BaseController):
    """Controller to handle buttons related to the conversion process"""

    def __init__(self, view: TurandotGtkView):
        self.msgqueue = multiprocessing.Queue()
        self.processor: Optional[ConversionProcessor] = None
        self.last_open_folder_id: Optional[int] = None
        BaseController.__init__(self, view)

    def connect_events(self):
        self.view.get_element("export_launch_button").connect("clicked", self._launch_conversion)
        self.view.get_element("export_cancel_button").connect("clicked", self._cancel_conversion)

    def _collect_job_settings(self) -> JobSettings:
        """Collect settings for conversion job from GUI"""
        convalg: ConversionAlgorithm = self.view.get_element("conversion_algorithm_dropdown").get_selected_value()
        refsrc: ReferenceSource = self.view.get_element("reference_source_dropdown").get_selected_value()
        if refsrc == ReferenceSource.ZOTERO:
            zotlib: Optional[int] = int(self.view.get_element("zotero_lib_dropdown").get_active_id())
        else:
            zotlib = None
        return JobSettings(convalg, refsrc, zotlib)

    @catch_exception
    def _collect_job(self) -> Optional[ConversionJob]:
        """Collect settings and assets for conversion job from GUI"""
        settings = self._collect_job_settings()
        srcstring = self.view.get_element("source_file_entry").get_text()
        if srcstring == "":
            raise TurandotAssetException('Field "Source file" must not be empty')
        srcasset = SourceAsset(path=Path(self.view.get_element("source_file_entry").get_text()), expand=True)
        tmplasset = self.view.get_element("template_dropdown").get_selected_asset()
        if settings.reference_source.value > 1:
            cslasset = self.view.get_element("csl_dropdown").get_selected_asset()
        else:
            cslasset = None
        if settings.reference_source == ReferenceSource.JSON:
            csljson = self.view.get_element("csljson_file_entry").get_text()
            if csljson == "":
                raise TurandotAssetException('Field "CSLJSON file" must not be empty')
        else:
            csljson = None
        assets = JobAssets(srcasset, tmplasset, cslasset, csljson)
        return ConversionJob(assets, settings, self.msgqueue)

    def _create_processor(self) -> Optional[ConversionProcessor]:
        """Create converter object"""
        job = self._collect_job()
        if job is None:
            return None
        chain = ConverterChain.build_chain(job.job_settings.conversion_algorithm)
        frontendstrat = GtkConversionUpdater(self.view)
        return ConversionProcessor(job, chain, frontendstrat)

    def _reattach_open_folder_callback(self, path: Path):
        """Attach correct path to callback on 'Open Folder' button"""
        if self.last_open_folder_id is not None:
            self.view.get_element("open_result_folder").disconnect(self.last_open_folder_id)
        self.last_open_folder_id = self.view.get_element("open_result_folder").connect("clicked", FrontendUtils.fm_open_path, path)

    def _launch_conversion(self, *args):
        """Launch conversion process"""
        self.processor = self._create_processor()
        if self.processor is None:
            return None
        self._reattach_open_folder_callback(self.processor.conversionjob.job_assets.sourcefile.directory)
        self.processor.start_conversion()

    def _cancel_conversion(self, *args):
        """Kill conversion process prematurely"""
        self.processor.kill_conversion()



from abc import ABC, abstractmethod
from enum import Enum
from typing import Type

from turandot.model import DatabaseAsset, TemplateAsset, CslAsset

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from turandot.ttk.presentations import DropdownObserver


class NotificationReason(Enum):
    """Enum to describe why the observer notifies the observable"""
    INIT = 0
    NEW = 1
    CHANGE = 2
    DELETE = 3


class DropdownObservable(ABC):
    """Base class for an observable to attach to a dropdown"""

    def __init__(self):
        self.observers: list["DropdownObserver"] = []

    def attach(self, observer: "DropdownObserver"):
        """Append an observer to the observable"""
        self.observers.append(observer)

    @abstractmethod
    def notify(self, reason: NotificationReason):
        """Notify the observers"""
        pass


class DatabaseDropdownObservable(DropdownObservable, ABC):
    """DB fed observable to attach comboboxes as observers to"""

    @abstractmethod
    def get_model(self) -> Type[DatabaseAsset]:
        """Return database model the observable manages"""
        pass

    def __init__(self):
        super(DatabaseDropdownObservable, self).__init__()
        self.entries: list[DatabaseAsset] = self._get_from_db()

    def _get_from_db(self) -> list[DatabaseAsset]:
        """Get entries from database to draw to dropdown"""
        model = self.get_model()
        entrylist = model.get_all(expand=True)
        entrylist.sort(key=lambda x: x.title.lower())
        return entrylist

    def notify(self, reason: NotificationReason):
        self.entries = self._get_from_db()
        for i in self.observers:
            i.update(self, reason=reason)


class TemplateObservable(DatabaseDropdownObservable):

    def get_model(self) -> Type[DatabaseAsset]:
        return TemplateAsset


class CslObservable(DatabaseDropdownObservable):

    def get_model(self) -> Type[DatabaseAsset]:
        return CslAsset



from enum import Enum
from md_citeproc import NotationStyle, OutputStyle
from turandot.model import ModelUtils
from turandot.ui import FrontendUtils
from turandot.gtk.controllers import BaseController
from turandot.gtk.presentations import SettingsControlBase, TitleControl, TextControl, SwitchControl, SpinControl, TextEnumControl


class SettingsController(BaseController):

    @staticmethod
    def _control_factory() -> list[SettingsControlBase]:
        """Build list of widgets to control application preferences"""
        return [
            TitleControl("General"),
            SwitchControl("Remember file input path:", ["general", "file_select_persistence"]),
            SwitchControl("Save intermediate files:", ["general", "save_intermediate"]),
            TitleControl("Zotero"),
            SpinControl("BetterBibtex Port:", ['api', 'zotero', 'port']),
            TitleControl("Table of contents"),
            TextControl("TOC Marker:", ['processors', 'convert_to_html', 'markdown_ext', 'toc', 'marker']),
            TitleControl("Citeproc"),
            TextControl("Locale:", ['processors', 'convert_to_html', 'markdown_ext', 'md_citeproc', 'locale']),
            TextEnumControl("Notation Style:",
                            ['processors', 'convert_to_html', 'markdown_ext', 'md_citeproc', 'notation'],
                            model=NotationStyle, text_based=True),
            TextEnumControl("Rendering Style:",
                            ['processors', 'convert_to_html', 'markdown_ext', 'md_citeproc', 'output'],
                            model=OutputStyle, text_based=True),
            TextControl("Footnotes Marker:",
                        ['processors', 'convert_to_html', 'markdown_ext', 'md_citeproc', 'footnotes_token']),
            TextControl("Bibliography Marker:",
                        ['processors', 'convert_to_html', 'markdown_ext', 'md_citeproc', 'bibliography_token']),
            TitleControl("Optional processors"),
            SwitchControl("Unified math markers:", ['opt_processors', 'unified_math_block_marker', 'enable']),
            SwitchControl("TOC pagination containers:", ['opt_processors', 'toc_pagination_containers', 'enable']),

        ]

    def _controls_to_view(self):
        """Build widget list and add to GUI"""
        clist = SettingsController._control_factory()
        for n, i in enumerate(clist, start=3):
            i.add_to_view(self.view, n)

    @staticmethod
    def _open_config_dir(*args):
        """Open config dir"""
        FrontendUtils.fm_open_path(ModelUtils.get_config_dir())

    def connect_events(self):
        self.view.get_element("config_file_location_entry").set_text(str(ModelUtils.get_config_file()))
        self.view.get_element("config_dir_open_button").connect("clicked", SettingsController._open_config_dir)
        self._controls_to_view()



import sys
import platform
from enum import Enum


class OpSys(Enum):
    LINUX = 0
    WINDOWS = 1
    MACOS = 2


class CpuArch(Enum):
    AMD64 = "amd64"


class SysInfo:
    """Get details about the platform the application is running on"""

    def __init__(self):
        self.system: OpSys = SysInfo._get_system()
        self.arch: CpuArch = SysInfo._get_arch()
        self.frozen: bool = SysInfo._get_frozen()

    @staticmethod
    def _get_system() -> OpSys:
        """Get OS as enum"""
        system = platform.system()
        if system == "Linux":
            return OpSys.LINUX
        if system == "Windows":
            return OpSys.WINDOWS
        if system == "Darwin":
            return OpSys.MACOS
        raise ValueError("Unsupported operating system: {}".format(system))

    @staticmethod
    def _get_arch() -> CpuArch:
        """Get architecture as enum"""
        machine = platform.machine()
        if machine in ['AMD64', 'x86_64']:
            return CpuArch.AMD64
        raise ValueError("Unsupported processor architecture: {}".format(machine))

    @staticmethod
    def _get_frozen() -> bool:
        """Check if application is frozen"""
        return getattr(sys, 'frozen', False)



from turandot.model import FrontendStrategy, CompanionData, QueueMessage, MessageType
from turandot.gtk import TurandotGtkView, guithread


class GtkConversionUpdater(FrontendStrategy):
    """GTK-specific implementation of the strategy to handle queue messages from the conversion process"""

    STATUS = {
        "idle": "IDLE",
        MessageType.EXCEPTION: "ERROR",
        MessageType.SUCCESS: "SUCCESS",
        MessageType.CANCELED: "CANCELED",
        "conv": "CONVERTING"
    }

    def __init__(self, view: TurandotGtkView):
        self.view = view
        self.warning_string: str = ""

    def handle_message(self, msg: QueueMessage):
        if msg.type == MessageType.STARTED:
            self._prime_gui(msg)
        elif msg.type == MessageType.NEXT_STEP:
            self._handle_next_step(msg)
        elif msg.type == MessageType.WARNING:
            self._handle_warning(msg)

    @guithread
    def handle_companion_data(self, data: CompanionData):
        self.view.get_element("general_status_label").set_text(
            self.STATUS[data.status.cause_of_death]
        )
        self.view.get_element("current_process_description_label").set_text("")
        if data.status.cause_of_death == MessageType.SUCCESS:
            self.view.get_element("steps_progress_bar").set_fraction(1.0)
            self.view.get_element("open_result_folder").set_visible(True)
        else:
            self.view.get_element("steps_progress_bar").set_fraction(0.0)
        if data.status.exception is not None:
            self.view.get_element("error_type_label").set_text(type(data.status.exception).__name__)
            self.view.get_element("error_text_buffer").set_text(
                "{}\n\n{}".format(str(data.status.exception), data.status.exception_tb)
            )
            self.view.get_element("error_feedback_container").set_visible(True)
        self.view.get_element("export_cancel_button").set_sensitive(False)
        self.view.get_element("export_launch_button").set_sensitive(True)

    @guithread
    def _prime_gui(self, msg: QueueMessage):
        """Reset GUI to default state at the beginning of a conversion"""
        self.view.get_element("general_status_label").set_text(self.STATUS["conv"])
        self.view.get_element("total_steps_label").set_text(str(msg.total_steps))
        self.view.get_element("n_step_label").set_text("0")
        self.view.get_element("steps_progress_bar").set_fraction(0.0)
        self.view.get_element("current_process_description_label").set_text("")
        self.view.get_element("warning_feedback_container").set_visible(False)
        self.warning_string: str = ""
        self.view.get_element("warning_text_buffer").set_text("")
        self.view.get_element("error_type_label").set_text("None")
        self.view.get_element("error_feedback_container").set_visible(False)
        self.view.get_element("error_text_buffer").set_text("")
        self.view.get_element("export_cancel_button").set_sensitive(True)
        self.view.get_element("export_launch_button").set_sensitive(False)
        self.view.get_element("open_result_folder").set_visible(False)

    @guithread
    def _handle_next_step(self, msg: QueueMessage):
        """Draw information about current conversion step to GUI"""
        self.view.get_element("n_step_label").set_text(str(msg.n_step))
        self.view.get_element("current_process_description_label").set_text(msg.step_desc)
        self.view.get_element("steps_progress_bar").set_fraction((msg.n_step -1) / msg.total_steps)

    @guithread
    def _handle_warning(self, msg: QueueMessage):
        """Draw warning message to GUI"""
        self.warning_string = "{}{}\n".format(self.warning_string, msg.warning)
        self.view.get_element("warning_text_buffer").set_text(self.warning_string)
        self.view.get_element("warning_feedback_container").set_visible(True)



import os
import shutil
import tempfile
from pathlib import Path
from peewee import DatabaseProxy, SqliteDatabase, Model, TextField
from turandot.meta import Singleton
from turandot.model import ModelUtils, ConfigModel


class DBInitializer(metaclass=Singleton):
    """Initialize SQLite database to hold templates & csl entries"""

    def __init__(self):
        self.proxy = DatabaseProxy()
        self.db = None

    # Copy initialization file if db does not exist, open db
    @staticmethod
    def _get_path() -> Path:
        """Locate SQLite file"""
        if ConfigModel().get_key(['debug', 'use_tmp_db']):
            dbpath = Path(tempfile.gettempdir()) / "turandot_assets.db"
            dbpath.unlink(missing_ok=True)
        else:
            dbpath = ModelUtils.get_config_dir() / "assets.db"
        return dbpath

    def get_db_proxy(self):
        """Use peewee proxy to init database on demand"""
        if self.db is None:
            self.db = SqliteDatabase(DBInitializer._get_path())
            self.proxy.initialize(self.db)
        return self.proxy


class BaseModel(Model):
    """Database object base model: Inherit to create tables"""

    class Meta:
        database = DBInitializer().get_db_proxy()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # slightly hacky approach to ensure table creation
        self._meta.database.create_tables([self])

    def set_data(self, **kwargs):
        """slightly hacky approach to write a dict to the db"""
        for i in self._meta.fields.keys():
            if i in kwargs.keys():
                val = kwargs[i]
                setattr(self, i, val)



import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from enum import Enum


class TypeFilterEnum(Enum):
    """Enumeration of different file type filters"""
    ALL = "all"
    MARKDOWN = "markdown"
    CSL = "csl"
    JSON = "json"
    HTML = "html"
    TMPL = "tmpl"


class FileTypeFilterFactory:
    """File ending based type filters for file picker dialogs"""

    @staticmethod
    def apply(t: TypeFilterEnum, dialog: Gtk.FileChooserDialog):
        """Apply filter to dialog"""
        lookup = {
            TypeFilterEnum.ALL: FileTypeFilterFactory._make_all(),
            TypeFilterEnum.MARKDOWN: FileTypeFilterFactory._make_markdown(),
            TypeFilterEnum.CSL: FileTypeFilterFactory._make_csl(),
            TypeFilterEnum.JSON: FileTypeFilterFactory._make_json(),
            TypeFilterEnum.HTML: FileTypeFilterFactory._make_html(),
            TypeFilterEnum.TMPL: FileTypeFilterFactory._make_tmpl()
        }
        if t != TypeFilterEnum.ALL:
            dialog.add_filter(lookup.get(t))
        dialog.add_filter(lookup.get(TypeFilterEnum.ALL))

    @staticmethod
    def _make_all() -> Gtk.FileFilter:
        """Create all files filter"""
        tf = Gtk.FileFilter()
        tf.set_name("All files")
        tf.add_pattern("*")
        return tf

    @staticmethod
    def _make_markdown() -> Gtk.FileFilter:
        """Create markdown file filters"""
        tf = Gtk.FileFilter()
        tf.set_name("Markdown files")
        tf.add_pattern("*.markdown")
        tf.add_pattern("*.mdown")
        tf.add_pattern("*.mkdn")
        tf.add_pattern("*.mkd")
        tf.add_pattern("*.md")
        tf.add_pattern("*.txt")
        return tf

    @staticmethod
    def _make_csl() -> Gtk.FileFilter:
        """Create csl file filter"""
        tf = Gtk.FileFilter()
        tf.set_name("CSL files")
        tf.add_pattern("*.csl")
        tf.add_pattern("*.xml")
        return tf

    @staticmethod
    def _make_json() -> Gtk.FileFilter:
        """Create json file filter"""
        tf = Gtk.FileFilter()
        tf.set_name("JSON files")
        tf.add_pattern("*.csljson")
        tf.add_pattern("*.json")
        return tf

    @staticmethod
    def _make_html() -> Gtk.FileFilter:
        """Create html file filter"""
        tf = Gtk.FileFilter()
        tf.set_name("HTML files")
        tf.add_pattern("*.html")
        tf.add_pattern("*.htm")
        tf.add_pattern("*.xml")
        return tf

    @staticmethod
    def _make_tmpl() -> Gtk.FileFilter:
        """Create template file filter"""
        tf = Gtk.FileFilter()
        tf.set_name("Template files")
        tf.add_pattern("*.tmpl")
        tf.add_pattern("*.zip")
        tf.add_pattern("*.yaml")
        return tf



from tkinter import W, E
from tkinter.messagebox import askyesno
from tkinter import filedialog
from typing import Optional, Type
from enum import Enum

from turandot.model import TemplateAsset, CslAsset
from turandot.ui import background
from turandot.ttk.filetypes import FileTypes
from turandot.ttk.view import TurandotTtkView
from turandot.ttk.view.styles import TTKStyles
from turandot.ttk.controllers import ControllerBase
from turandot.ttk.presentations import \
    DbFedFilePicker, \
    NotificationReason, \
    DatabaseDropdownObservable, \
    TemplateObservable, \
    CslObservable, \
    DatabaseDropdown, \
    DatabaseNewOptDropdown


class DbDropdownController(ControllerBase):
    """Draw comboboxes attached to db fed observables"""

    def __init__(self, view: TurandotTtkView):
        super().__init__(view)

        # Create observables
        self.template_subject = TemplateObservable()
        self.csl_subject = CslObservable()
        # Create Template dropdown on "Converter" Tab
        self.template_selector_dropdown = self._create_dropdown(
            "converter_template_dropdown_frame", "converter_template_dropdown", self.template_subject, DatabaseDropdown
        )
        # Create Template dropdown on "Templates" Tab
        self.template_editor_dropdown = self._create_dropdown(
            "template_editor_dropdown_frame", "template_editor_dropdown", self.template_subject, DatabaseNewOptDropdown
        )
        # Create CSL dropdown on "Converter" Tab
        self.csl_selector_dropdown = self._create_dropdown(
            "converter_csl_dropdown_frame", "converter_csl_dropdown", self.csl_subject, DatabaseDropdown
        )
        # Create CSL dropdown on "CSL" Tab
        self.csl_editor_dropdown = self._create_dropdown(
            "csl_editor_dropdown_frame", "csl_editor_dropdown", self.csl_subject, DatabaseNewOptDropdown
        )
        self._attach_template_tab_events()
        self._attach_csl_tab_events()

    def _create_dropdown(
            self, add_to: str, widget_key: str, observable: DatabaseDropdownObservable, cls: Type[DatabaseDropdown]
    ) -> DatabaseDropdown:
        """Create instance of a combobox with an attached observable"""
        add_to = self.view.widgets.get(add_to)
        self.view.widgets[widget_key] = drpdwn = cls(add_to, state="readonly")
        observable.attach(drpdwn)
        drpdwn.update(observable, reason=NotificationReason.INIT)
        drpdwn.grid(row=0, column=0, sticky=(W, E), **TTKStyles.get_padding().right_entries)
        return drpdwn

    def _attach_template_tab_events(self):
        """Attach callbacks for the template tab"""
        self.template_editor_dropdown.set_callback(self._template_tab_select_callback)
        self.view.widgets["tmpl_delete_button"].bind("<Button-1>", self._template_delete_callback)
        self.view.widgets["tmpl_save_button"].bind("<Button-1>", self._template_save_callback)
        self.view.widgets["tmpl_base_select_button"].bind("<Button-1>", self._template_picker_callback)

    def _attach_csl_tab_events(self):
        """Attach callback for the csl tab"""
        self.csl_editor_dropdown.set_callback(self._csl_tab_select_callback)
        self.view.widgets["csl_save_button"].bind("<Button-1>", self._csl_tab_save_callback)
        self.view.widgets["csl_delete_button"].bind("<Button-1>", self._csl_delete_callback)
        self.view.widgets["csl_select_button"].bind("<Button-1>", self._csl_file_picker)

    def _template_tab_select_callback(self, *args):
        """Attach callback on template selection"""
        dbid = self.template_editor_dropdown.get()
        if dbid == 0:
            self.view.widgets["tmpl_base_file_enty_value"].set("")
            self.view.widgets["tmpl_allow_jinja_value"].set(False)
            self.view.widgets["tmpl_allow_mako_value"].set(False)
            self.view.widgets["tmpl_delete_button"].state(["disabled"])
            return
        tmpl = TemplateAsset.get(dbid=dbid, expand=True)
        self.view.widgets["tmpl_base_file_enty_value"].set(tmpl.path)
        self.view.widgets["tmpl_allow_jinja_value"].set(tmpl.allow_jinja)
        self.view.widgets["tmpl_allow_mako_value"].set(tmpl.allow_mako)
        self.view.widgets["tmpl_delete_button"].state(["!disabled"])

    @background
    def _template_save_callback(self, *args):
        """Attach callback to template save button"""
        dbid = self.template_editor_dropdown.get()
        if dbid == 0:
            tmpl = TemplateAsset(
                path=self.view.widgets["tmpl_base_file_enty_value"].get(),
                allow_jinja=self.view.widgets["tmpl_allow_jinja_value"].get(),
                allow_mako=self.view.widgets["tmpl_allow_mako_value"].get()
            )
            tmpl.save()
            self.template_subject.notify(reason=NotificationReason.NEW)
        else:
            tmpl = TemplateAsset.get(dbid=dbid)
            tmpl.path = self.view.widgets["tmpl_base_file_enty_value"].get()
            tmpl.allow_jinja = self.view.widgets["tmpl_allow_jinja_value"].get()
            tmpl.allow_mako = self.view.widgets["tmpl_allow_mako_value"].get()
            tmpl.save()
            self.template_subject.notify(reason=NotificationReason.CHANGE)

    @background
    def _template_delete_callback(self, *args):
        """Attach callback to template delete button"""
        answer = askyesno("Confirm", "Remove template entry?\n(No files will be deleted)")
        if answer:
            dbid = self.template_editor_dropdown.get()
            tmpl = TemplateAsset.get(dbid=dbid)
            tmpl.delete()
            self.template_subject.notify(reason=NotificationReason.DELETE)

    @background
    def _template_picker_callback(self, *args):
        """Draw file select dialog to pick a template"""
        DbFedFilePicker.draw(self.view, "tmpl_base_file_enty_value", FileTypes.templates)

    def _csl_tab_select_callback(self, *args):
        dbid = self.csl_editor_dropdown.get()
        if dbid == 0:
            self.view.widgets["csl_base_file_entry_value"].set("")
            self.view.widgets["csl_delete_button"].state(["disabled"])
        else:
            csl = CslAsset.get(dbid=dbid)
            self.view.widgets["csl_base_file_entry_value"].set(csl.path)
            self.view.widgets["csl_delete_button"].state(["!disabled"])

    @background
    def _csl_tab_save_callback(self, *args):
        """Attach callback to csl save button"""
        dbid = self.csl_editor_dropdown.get()
        if dbid == 0:
            csl = CslAsset(path=self.view.widgets["csl_base_file_entry_value"].get())
            csl.save()
            self.csl_subject.notify(reason=NotificationReason.NEW)
        else:
            csl = CslAsset.get(dbid=dbid)
            csl.path = self.view.widgets["csl_base_file_entry_value"].get()
            csl.save()
            self.csl_subject.notify(reason=NotificationReason.CHANGE)

    @background
    def _csl_delete_callback(self, *args):
        """Attach callback to csl delete button"""
        answer = askyesno("Confirm", "Remove csl entry?\n(No files will be deleted)")
        if answer:
            dbid = self.csl_editor_dropdown.get()
            csl = CslAsset.get(dbid=dbid)
            csl.delete()
            self.csl_subject.notify(reason=NotificationReason.DELETE)

    @background
    def _csl_file_picker(self, *args):
        """Draw file select dialog to pick a csl file"""
        DbFedFilePicker.draw(self.view, "csl_base_file_entry_value", FileTypes.csl)



from turandot.ttk.controllers.basecontroller import ControllerBase
from turandot.ttk.controllers.settingscontroller import SettingsController
from turandot.ttk.controllers.dbdropdowncontroller import DbDropdownController
from turandot.ttk.controllers.convertercontroller import ConverterController
from turandot.ttk.controllers.exceptioncontroller import ExceptionController
from turandot.ttk.controllers.aboutcontroller import AboutController
from turandot.ttk.controllers.conversionupdater import TtkConversionUpdater
from turandot.ttk.controllers.exportcontroller import ExportController



from turandot import TurandotConversionException
from turandot.model import ConverterBase, GatherData, CopyTemplate, ConversionAlgorithm, ConvertToHtml, OptionalStage, ApplyTemplate, WeasyprintToPdf
from turandot.model.converter.optional import UnifiedMathMarker, TocPaginationContainers


class ConverterChain:

    @staticmethod
    def build_chain(alg: ConversionAlgorithm) -> ConverterBase:
        """Build a chain of converter for a specified algorithm"""
        if alg == ConversionAlgorithm.WEASYPRINT:
            return ConverterChain._build_weasyprint()
        raise TurandotConversionException("Unknown conversion algorithm")

    @staticmethod
    def _build_weasyprint() -> ConverterBase:
        """Build Weasyprint PDF conversion algorithm chain"""
        chain = GatherData()
        chain.chain_append(CopyTemplate())
        html = ConvertToHtml()
        html.register_optional(UnifiedMathMarker(), OptionalStage.PRE)
        chain.chain_append(html)
        tmpl = ApplyTemplate()
        tmpl.register_optional(TocPaginationContainers(), OptionalStage.POST)
        chain.chain_append(tmpl)
        chain.chain_append(WeasyprintToPdf())
        return chain



from pathlib import Path
from tkinter import filedialog

from turandot.model import FileSelectPersistApi

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from turandot.ttk.view import TurandotTtkView


class DbFedFilePicker:
    """Util class to draw a file picker dialog"""

    @staticmethod
    def draw(view: "TurandotTtkView", return_id: str, filefilter: list[tuple[str]], title="Choose a file"):
        """Draw file picker dialog"""
        initialdir = FileSelectPersistApi.get(return_id)
        chosen = filedialog.askopenfilename(filetypes=filefilter, title=title, initialdir=initialdir)
        if chosen not in [(), ""]:
            FileSelectPersistApi.set(return_id, Path(chosen).parent)
            view.widgets[return_id].set(chosen)



import os
from pathlib import Path
from importlib import resources


class ModelUtils:
    """Static helper functions for the model"""

    @staticmethod
    def get_config_dir() -> Path:
        # Config path for Windows, something like C:\User\username\Appdata\Roaming
        if 'APPDATA' in os.environ:
            confighome = Path(os.environ['APPDATA']).resolve()
        # Config path for macOS
        elif 'XDG_CONFIG_HOME' in os.environ:
            confighome = Path(os.environ['XDG_CONFIG_HOME']).resolve()
        # Config path for Linux, /home/username/.config
        else:
            confighome = Path(os.environ['HOME']).resolve() / '.config'
        configpath = confighome / 'turandot'
        return configpath

    @staticmethod
    def get_config_file() -> Path:
        return ModelUtils.get_config_dir() / "config3.yaml"

    @staticmethod
    def get_asset_content(filename: str) -> str:
        return resources.read_text("turandot.assets", filename)

    @staticmethod
    def get_asset_bytes(filename: str) -> bytes:
        return resources.read_binary("turandot.assets", filename)



import os
import subprocess
from pathlib import Path
from turandot import SysInfo, OpSys


class FrontendUtils:
    """Static utilities for frontends independent of their framework"""

    @staticmethod
    def fm_open_path(*args):
        """Open path in systems file manager"""
        d = None
        for i in args:
            if isinstance(i, Path):
                d = i
                break
        if d is None:
            raise ValueError("No path parameter in func call")
        sysinfo = SysInfo()
        if sysinfo.system == OpSys.WINDOWS:
            os.startfile(d, "explore")
        elif sysinfo.system == OpSys.MACOS:
            subprocess.Popen(['open', d])
        else:
            subprocess.Popen(['xdg-open', d])

    @staticmethod
    def replace_version_number(asset_text: str) -> str:
        """Replace module version number in 'about' tab"""
        from turandot import __version__ as v
        return asset_text.replace("{VER}", v)



from pathlib import Path
from peewee import DoesNotExist
from turandot.model.db.dbmodels import FileSelectPersistence
from turandot.model import ConfigModel


class FileSelectPersistApi:
    """Get/set recently used paths for file selector dialogs"""

    @staticmethod
    def get(selector_id: str) -> Path:
        """Get recently used path for a specific GUI element"""
        if ConfigModel().get_key(["general", "file_select_persistence"], False):
            try:
                entry = FileSelectPersistence.get(input_id=selector_id)
                return Path(entry.last_path)
            except DoesNotExist:
                pass
        return Path.home()

    @staticmethod
    def set(selector_id: str, path: Path):
        """Set recently used path for a specific GUI element"""
        try:
            entry = FileSelectPersistence.get(input_id=selector_id)
            entry.last_path = path
        except DoesNotExist:
            entry = FileSelectPersistence.create(input_id=selector_id, last_path=path)
        entry.save()



from tkinter import *
from tkinter import ttk

from turandot.ttk.view import TTKStyles
from turandot.ttk.view import ViewComponent


class ProcessFrame(ViewComponent):
    """Create view component to monitor process"""

    def _create_status_labels(self, pframe: ttk.Frame):
        ttk.Label(pframe, text="Status:", style="h2.TLabel", padding=TTKStyles.get_padding().h2_top)\
            .grid(row=0, column=0, sticky=W)
        self.widgets["status_label_value"] = StringVar(value="IDLE")
        ttk.Label(pframe, textvariable=self.widgets["status_label_value"]).grid(row=0, column=1, sticky=E)

    def _create_step_bar(self, pframe: ttk.Frame):
        ttk.Label(pframe, text="Step:").grid(
            row=2, column=0, sticky=W, **TTKStyles.get_padding().left_labels
        )
        n_steps_frame = ttk.Frame(pframe)
        n_steps_frame.grid(
            row=2, column=1, sticky=E, **TTKStyles.get_padding().right_entries
        )
        self.widgets["n_steps_label_value"] = StringVar(value=0)
        ttk.Label(n_steps_frame, textvariable=self.widgets["n_steps_label_value"]).grid(row=0, column=0, sticky=E)
        ttk.Label(n_steps_frame, text="/").grid(row=0, column=1, sticky=E)
        self.widgets["total_steps_label_value"] = StringVar(value=0)
        ttk.Label(n_steps_frame, textvariable=self.widgets["total_steps_label_value"]).grid(row=0, column=2, sticky=E)
        self.widgets["steps_progress_bar"] = ttk.Progressbar(pframe, length=300, mode="determinate", orient=HORIZONTAL)
        self.widgets["steps_progress_bar"].grid(
            row=3, columnspan=2, sticky=(W, E), **TTKStyles.get_padding().full_process_container
        )
        self.widgets["steps_progress_bar"]["value"] = 0
        ttk.Label(pframe, text="Process:").grid(
            row=4, column=0, sticky=W, **TTKStyles.get_padding().left_labels
        )
        self.widgets["current_process_description_label_value"] = StringVar(value="idle")
        ttk.Label(pframe, textvariable=self.widgets["current_process_description_label_value"]).grid(
            row=4, column=1, sticky=E
        )

    def _create_warning_widgets(self, pframe: ttk.Frame):
        self.widgets["warnings_title"] = ttk.Label(pframe, text="Warnings", style="h2.TLabel", padding=TTKStyles.get_padding().h2)
        self.widgets["warnings_title"].grid(row=5, columnspan=2, sticky=(W, E))
        self.widgets["warnings_title"].grid_remove()
        self.widgets["warning_text_view"] = Text(pframe, height=10, width=40)
        self.widgets["warning_text_view"].grid(
            row=6, columnspan=2, sticky=(W, E), **TTKStyles.get_padding().full_process_container
        )
        self.widgets["warning_text_view"].grid_remove()

    def _create_error_widgets(self, pframe: ttk.Frame):
        self.widgets["error_title"] = ttk.Label(pframe, text="Error/Traceback", style="h2.TLabel", padding=TTKStyles.get_padding().h2)
        self.widgets["error_title"].grid(row=7, columnspan=2, sticky=(W, E))
        self.widgets["error_title"].grid_remove()
        self.widgets["error_type_desc_label"] = ttk.Label(pframe, text="Type:")
        self.widgets["error_type_desc_label"].grid(
            row=8, column=0, sticky=W, **TTKStyles.get_padding().left_labels
        )
        self.widgets["error_type_desc_label"].grid_remove()
        self.widgets["error_type_label_value"] = StringVar(value="None")
        self.widgets["error_type_label"] = ttk.Label(pframe, textvariable=self.widgets["error_type_label_value"])
        self.widgets["error_type_label"].grid(row=8, column=1, sticky=E)
        self.widgets["error_type_label"].grid_remove()
        self.widgets["error_text_view"] = Text(pframe, height=10, width=40)
        self.widgets["error_text_view"].grid(
            row=9, columnspan=2, sticky=(W, E), **TTKStyles.get_padding().full_process_container
        )
        self.widgets["error_text_view"].grid_remove()

    def _create_open_button(self, pframe: ttk.Frame):
        self.widgets["open_result_folder"] = ttk.Button(pframe, text="Open folder")
        self.widgets["open_result_folder"].grid(
            row=10, columnspan=2, sticky=E, **TTKStyles.get_padding().right_entries
        )
        self.widgets["open_result_folder"].grid_remove()

    def create(self, add_to: ttk.Frame) -> ttk.Frame:
        pframe = ttk.Frame(add_to)
        pframe.grid(column=0, row=0, sticky=(N, S, W, E))
        pframe.columnconfigure(0, weight=1)
        pframe.columnconfigure(1, weight=1)
        self._create_status_labels(pframe)
        self._create_step_bar(pframe)
        self._create_warning_widgets(pframe)
        self._create_error_widgets(pframe)
        self._create_open_button(pframe)
        return pframe



from abc import ABC, abstractmethod
from enum import Enum, EnumMeta
from typing import Union
from gi.repository import Gtk
from typing import TYPE_CHECKING
from turandot.model import ConfigModel
from turandot.gtk.presentations import EnumDropdown
if TYPE_CHECKING:
    from turandot.gtk import TurandotGtkView


class SettingsViewBase(ABC):
    """Base class for read-only settings widgets"""

    def __init__(self, label: str):
        self.label = self._create_label(label)

    @abstractmethod
    def _create_label(self, label: str) -> Gtk.Label:
        """Create text label for the configuration control widget"""
        pass

    @abstractmethod
    def add_to_view(self, view: "TurandotGtkView", row: int):
        """Add widget elements to GUI"""
        pass


class SettingsControlBase(SettingsViewBase, ABC):
    """Base class for savable widgets"""

    def __init__(self, label, config_key: Union[str, list]):
        super().__init__(label)
        self.config_key = config_key

    def _create_label(self, label: str) -> Gtk.Label:
        lbl = Gtk.Label(label)
        lbl.set_halign(Gtk.Align.START)
        lbl.show()
        return lbl

    @abstractmethod
    def _attach_callback(self):
        """Attach callback method to specific change event"""
        pass

    @abstractmethod
    def _callback(self):
        """Method to call on change"""
        pass


class TitleControl(SettingsViewBase):
    """Title label, read only"""

    def __init__(self, title: str, vmargins: tuple = (20, 10)):
        super().__init__(title)
        self.label.set_margin_top(vmargins[0])
        self.label.set_margin_bottom(vmargins[1])

    def _create_label(self, label: str) -> Gtk.Label:
        lbl = Gtk.Label()
        lbl.set_markup("<b>" + label + "</b>")
        lbl.set_hexpand(True)
        lbl.set_halign(Gtk.Align.START)
        lbl.show()
        return lbl

    def add_to_view(self, view: "TurandotGtkView", row: int):
        view.get_element("dynamic_settings_grid").attach(self.label, 0, row, 2, 1)


class TextControl(SettingsControlBase):
    """Control widget to set strings in config file"""

    def __init__(self, label: str, config_key: Union[str, list]):
        super().__init__(label, config_key)
        self.entry = Gtk.Entry()
        self.entry.set_text(ConfigModel().get_key(self.config_key, default=""))
        self.entry.set_hexpand(True)
        self.entry.show()
        self._attach_callback()

    def _callback(self, *args):
        ConfigModel().set_key(self.config_key, self.entry.get_text())

    def _attach_callback(self):
        self.entry.connect("changed", self._callback)

    def add_to_view(self, view: "TurandotGtkView", row: int):
        view.get_element("dynamic_settings_grid").attach(self.label, 0, row, 1, 1)
        view.get_element("dynamic_settings_grid").attach(self.entry, 1, row, 1, 1)


class SwitchControl(SettingsControlBase):
    """Control widget to set bools in config file"""

    def __init__(self, label: str, config_key: Union[str, list]):
        super().__init__(label, config_key)
        self.switch = Gtk.Switch()
        self.switch.set_active(ConfigModel().get_key(self.config_key, default=False))
        self.switch.set_hexpand(True)
        self.switch.set_halign(Gtk.Align.END)
        self.switch.show()
        self._attach_callback()

    def _attach_callback(self):
        self.switch.connect("notify::active", self._callback)

    def _callback(self, *args):
        ConfigModel().set_key(self.config_key, self.switch.get_active())

    def add_to_view(self, view: "TurandotGtkView", row: int):
        view.get_element("dynamic_settings_grid").attach(self.label, 0, row, 1, 1)
        view.get_element("dynamic_settings_grid").attach(self.switch, 1, row, 1, 1)


class SpinControl(SettingsControlBase):
    """Control widget to set ints in config file"""

    def __init__(
            self,
            label: str,
            config_key: Union[str, list],
            lower: float = 1024,
            upper: float = 65553,
            step: float = 1,
            decimal: int = 0
    ):
        super().__init__(label, config_key)
        self.decimal = decimal
        adjustment = Gtk.Adjustment(lower=lower, upper=upper, step_increment=step)
        self.spinner = Gtk.SpinButton()
        self.spinner.set_adjustment(adjustment)
        self.spinner.set_numeric(True)
        self.spinner.set_digits(decimal)
        self.spinner.set_update_policy(Gtk.SpinButtonUpdatePolicy.ALWAYS)
        self.spinner.set_value(ConfigModel().get_key(self.config_key, default=False))
        self.spinner.show()
        self._attach_callback()

    def _attach_callback(self):
        self.spinner.connect("value-changed", self._callback)
        self.spinner.connect("scroll-event", self._empty_callback)

    def _callback(self, *args):
        if self.decimal <= 0:
            val = self.spinner.get_value_as_int()
        else:
            val = self.spinner.get_value()
        ConfigModel().set_key(self.config_key, val)

    @staticmethod
    def _empty_callback(*args):
        """empty callback to prevent Spinners from changing on scrolling"""
        return True

    def add_to_view(self, view: "TurandotGtkView", row: int):
        view.get_element("dynamic_settings_grid").attach(self.label, 0, row, 1, 1)
        view.get_element("dynamic_settings_grid").attach(self.spinner, 1, row, 1, 1)


class TextEnumControl(SettingsControlBase):
    """Control widget to set enum values in config file"""

    def __init__(self, label: str, config_key: Union[list, str], model: EnumMeta, text_based: bool = False):
        super().__init__(label, config_key)
        self.combo = EnumDropdown(model, text_based=text_based)
        initval = ConfigModel().get_key(self.config_key, default="")
        self.combo.set_selected_value(initval)
        self.combo.show()
        self._attach_callback()

    def _attach_callback(self):
        self.combo.connect("changed", self._callback)

    def _callback(self, *args):
        val = self.combo.get_selected_value()
        ConfigModel().set_key(self.config_key, val.value)

    def add_to_view(self, view: "TurandotGtkView", row: int):
        view.get_element("dynamic_settings_grid").attach(self.label, 0, row, 1, 1)
        view.get_element("dynamic_settings_grid").attach(self.combo, 1, row, 1, 1)



import os
from turandot import SysInfo, OpSys
from turandot.model import ModelUtils
from turandot.gtk.controllers import BaseController, DialogController, DataDropdownController, SettingsController, ConverterTabController, AboutController, ExportController
from turandot.ui import ExceptionCatcher, TurandotFrontend
from turandot.gtk import GtkCatcher, TurandotGtkView


class TurandotGtk(TurandotFrontend):
    """Launch GTK GUI for Turandot"""

    def __init__(self):
        # Create view
        self.view = TurandotGtkView()
        # Create concrete exception catcher and pass the view to it
        concrete_catcher = GtkCatcher(self.view)
        self.catcher = ExceptionCatcher()
        self.catcher.set_strategy(concrete_catcher)
        self.controllers: list[BaseController] = [
            DialogController(self.view),
            DataDropdownController(self.view),
            SettingsController(self.view),
            ConverterTabController(self.view),
            AboutController(self.view),
            ExportController(self.view)
        ]
        for i in self.controllers:
            i.connect_events()

    def run(self):
        """Run application"""
        self.view.show_main()


if __name__ == "__main__":
    TurandotGtk().run()



import re
from turandot.model import OptionalConverter, ConversionJob, ConfigDict


class UnifiedMathMarker(OptionalConverter):
    """Optional processor to convert JS style Katex fences to Python Markdown compatible fences"""

    LATEX_PATTERN = '(^\$\$$)([^\$]+)(^\$\$$)'
    GITHUB_MARKERS = ['```math', '```']

    def check_config(self, config: dict) -> bool:
        return bool(
            self.conversion_job.config.get_key(['opt_processors', 'unified_math_block_marker', 'enable'], default=False)
        )

    def process_step(self) -> ConversionJob:
        self.conversion_job.current_step.content = UnifiedMathMarker._static_conversion(self.conversion_job.current_step.content)
        return self.conversion_job

    @staticmethod
    def _static_conversion(fulltext: str) -> str:
        """Find regex matches for JS Katex fences and convert them"""
        flagged_pattern = re.compile('(^\$\$$)([^\$]+)(^\$\$$)', flags=re.MULTILINE)
        while True:
            stack = re.finditer(flagged_pattern, fulltext)
            if (needle := next(stack, None)) is None:
                break
            fulltext = UnifiedMathMarker._process_latex_match(needle, fulltext)
        return fulltext

    @staticmethod
    def _process_latex_match(match: re.Match, fulltext: str) -> str:
        """Process a found fence match"""
        pre = fulltext[:match.start()]
        math = fulltext[match.start() + 2:match.end() - 2]
        post = fulltext[match.end():]
        return pre + UnifiedMathMarker.GITHUB_MARKERS[0] + math + UnifiedMathMarker.GITHUB_MARKERS[1] + post



from dataclasses import dataclass


class FileTypesShared:

    all = [("All files", "*")]


class FileTypes(FileTypesShared):
    """Constant class for all file type filters to attach to picker dialogs"""

    markdown = [("Markdown files", ".markdown .mdown .mkdn .mkd .md .txt"), FileTypesShared.all[0]]
    csl = [("CSL files", ".csl .xml"), FileTypesShared.all[0]]
    json = [("JSON files", ".csljson .json"), FileTypesShared.all[0]]
    html = [("HTML files", ".html .html .xml"), FileTypesShared.all[0]]
    templates = [("Template files", ".tmpl .zip .yaml"), FileTypesShared.all[0]]



from abc import ABC, abstractmethod
from typing import Optional
import traceback
from functools import wraps

from turandot import Singleton


class CatcherStrategy(ABC):

    @abstractmethod
    def handle_exception(self, e: Exception, tb: str):
        """Forward caught exception to GUI implementation"""
        pass


class ExceptionCatcher(metaclass=Singleton):
    """Catch exception and forward to strategy, needs to be instanced by GUI first!"""

    def __init__(self):
        self.strat: Optional[CatcherStrategy] = None

    def set_strategy(self, strat: CatcherStrategy):
        """Define GUI-specific strategy to output exceptions"""
        self.strat = strat

    def catch(self, e: Exception, tb: str):
        """Pass exception to specific GUI implementation"""
        if self.strat is None:
            raise ValueError("Set exception catcher strategy first!")
        self.strat.handle_exception(e, tb)


def catch_exception(f):
    """Decorator: Catch exception, show on frontend"""
    @wraps(f)
    def catch(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            tb = traceback.format_exc()
            catcher = ExceptionCatcher()
            catcher.catch(e, tb)
    return catch



from abc import ABC, abstractmethod
from tkinter import *
from tkinter import ttk


class ViewComponent(ABC):
    """Base class from view components"""

    def __init__(self):
        self.widgets = {}

    @abstractmethod
    def create(self, add_to: ttk.Frame) -> ttk.Frame:
        """Create and add the components widgets to the view"""
        pass



import tempfile
import uuid
from copy import copy
from pathlib import Path
from turandot import TurandotAssetException
from turandot.model import ConverterBase, TemplateAsset, CslAsset, TextAsset, SourceAsset, ConversionJob, ReferenceSource, QueueMessage, ZoteroConnector


class GatherData(ConverterBase):
    """Get all data files necessary for the conversion process, complete ConversionJob object"""

    DESC = "Gathering data"

    def _config_interlace(self):
        """
        Override application config with values from document and template
        document takes priority over template
        template takes priority over application
        """
        tmplconfig = self.conversion_job.job_assets.template.metadata.get("turandot_config", {})
        sourceconfig = self.conversion_job.job_assets.sourcefile.metadata.get("turandot_config", {})
        if not isinstance(tmplconfig, dict):
            tmplconfig = {}
        if not isinstance(sourceconfig, dict):
            sourceconfig = {}
        self.conversion_job.config = self.conversion_job.config.interlace(tmplconfig)
        self.conversion_job.config = self.conversion_job.config.interlace(sourceconfig)

    def _force_expand(self):
        """Create expanded asset objects from asset paths"""
        if isinstance(self.conversion_job.job_assets.sourcefile, (str, Path)):
            source = SourceAsset(Path(self.conversion_job.job_assets.sourcefile), expand=True)
            self.conversion_job.job_assets.sourcefile = source
        if isinstance(self.conversion_job.job_assets.template, int):
            tmpl = TemplateAsset.get(dbid=self.conversion_job.job_assets.template, expand=True)
            self.conversion_job.job_assets.template = tmpl
        if isinstance(self.conversion_job.job_assets.cslfile, int):
            csl = CslAsset.get(dbid=self.conversion_job.job_assets.cslfile, expand=True)
            self.conversion_job.job_assets.cslfile = csl
        # Create CSLJSON asset from Zotero lib if necessary
        if self.conversion_job.job_settings.reference_source == ReferenceSource.ZOTERO:
            csljson_drop = self._zotero_to_file(self.conversion_job.job_settings.zotero_lib_id)
            # Log file drop to delete on cleanup
            self.conversion_job.msgqueue.put(QueueMessage.copymsg((csljson_drop, 'f')))
            self.conversion_job.job_assets.csljson = csljson_drop
        if isinstance(self.conversion_job.job_assets.csljson, (str, Path)):
            csljson = TextAsset(Path(self.conversion_job.job_assets.csljson), expand=True)
            self.conversion_job.job_assets.csljson = csljson

    def _zotero_to_file(self, lib_id: int) -> Path:
        """Get CSLJSON data from Zotero & write to a file"""
        filename = "zotero_lib_{}_{}.json".format(lib_id, str(uuid.uuid4()))
        drop_file = Path(tempfile.gettempdir()) / filename
        with drop_file.open('w') as fh:
            fh.write(
                ZoteroConnector(self.conversion_job.config).get_csljson(libid=lib_id)
            )
        return drop_file

    def _check_exceptions(self):
        """Check expansion processes for exceptions: Did anything go wrong while reading files?"""
        for i in self.conversion_job.job_assets:
            if getattr(i, "exception", None) is not None:
                raise i.exception
        if self.conversion_job.job_assets.template is None:
            raise TurandotAssetException("A template for the conversion process must be specified.")
        if self.conversion_job.job_settings.reference_source.value >= ReferenceSource.ZOTERO.value and self.conversion_job.job_assets.cslfile is None:
            raise TurandotAssetException("With the selected reference data source, a citation style must be specified.")

    def process_step(self) -> ConversionJob:
        self._force_expand()
        self._check_exceptions()
        self._config_interlace()
        self.conversion_job.current_step.content = copy(self.conversion_job.job_assets.sourcefile.content)
        return self.conversion_job



from abc import ABC, abstractmethod
from typing import Optional, Union
from pathlib import Path
import peewee
from turandot.model.datastructures.baseasset import BaseAsset
from turandot.model.db.dbinit import BaseModel


class DatabaseAsset(BaseAsset, ABC):
    """Base class for conversion assets that are savable to and loadable from the database"""

    DB_MODEL = BaseModel
    ZERO_VALUES = [None, "0", 0]
    TITLE_NOT_FOUND_MSG = "- title not found -"

    def __init__(self, path: Union[str, Path], dbid: Optional[int] = None, expand: bool = False):
        self.dbid = dbid
        self.title = DatabaseAsset.TITLE_NOT_FOUND_MSG
        BaseAsset.__init__(self, path=path, expand=expand)

    @abstractmethod
    def _read_title(self):
        """Read title from database asset"""
        pass

    @abstractmethod
    def _to_orm(self) -> BaseModel:
        """Convert asset data to ORM model"""
        pass

    @classmethod
    @abstractmethod
    def _from_orm(cls, model: BaseModel, expand: bool):
        """Create conversion asset from database entry"""
        pass

    @classmethod
    def get(cls, dbid: int, expand: bool = False):
        """Get conversion asset by database id"""
        try:
            entry = cls.DB_MODEL.get(dbid=dbid)
        except peewee.DoesNotExist:
            return None
        return cls._from_orm(entry, expand)

    @classmethod
    def get_all(cls, expand: bool = False) -> list:
        """Get list of all database assets of this specific class"""
        objls = []
        res = cls.DB_MODEL.select().where(1)
        for i in res:
            objls.append(cls._from_orm(i, expand))
        objls.sort(key=lambda x: x.title.lower())
        return objls

    def save(self):
        """Save conversion asset to database"""
        entry = self._to_orm()
        entry.save()
        self.dbid = entry.dbid

    def delete(self):
        """Delete conversion asset from the database"""
        if self.dbid is None:
            return None
        entry = self.get(self.dbid)
        if entry is not None:
            self._to_orm().delete_instance()



from typing import Any, Callable
from bidict import bidict
from tkinter import ttk
import tkinter as tk


class KvCombobox(ttk.Combobox):
    """Key-value combobox to return something else than strings"""

    def __init__(self, master, **kwargs):
        kwargs["values"] = ()
        kwargs.pop("textvariable", None)
        self._sv = tk.StringVar()
        super(KvCombobox, self).__init__(master=master, textvariable=self._sv, **kwargs)
        self._lookup = bidict({})
        self._optstrings = []

    def clear(self):
        """Remove all options from the combobox"""
        self._lookup = bidict({})
        self._optstrings = []
        self["values"] = ()

    def _unique_option(self, option: str) -> str:
        """Create a new unique value for a given key"""
        while option in self._optstrings:
            option = f"{option}*"
        return option

    def add_option(self, key: Any, stringval: str):
        """Add new key-value pair to combobox"""
        unique_string = self._unique_option(stringval)
        self._optstrings.append(unique_string)
        self._lookup[key] = unique_string
        vallist = []
        for i in self._lookup.values():
            vallist.append(i)
        self["values"] = vallist

    def get(self) -> Any:
        """Get selected key from combobox"""
        textval = self._sv.get()
        return self._lookup.inverse.get(textval)

    def set(self, key: Any):
        """Set selected key"""
        textval = self._lookup.get(key)
        self._sv.set(textval)

    def set_callback(self, f: Callable):
        """Attach callback function to combobox"""
        self._sv.trace_add("write", f)

    def check_key(self, key: int) -> bool:
        """Check if key is in combobox"""
        return key in self._lookup.keys()

    def get_keys(self) -> list:
        """Return list of keys present in the combobox"""
        return list(self._lookup.keys())



from typing import Optional
from abc import ABC, abstractmethod
from turandot.model import ConversionJob, OptionalStage, QueueMessage


class AbstractConverter(ABC):
    """Base class for optional & mandatory converters, specifies basic interface"""

    def __init__(self):
        self.conversion_job: Optional[ConversionJob] = None

    @abstractmethod
    def handle(self, conversion_job: ConversionJob):
        """Handle a conversion job by processing the current conversion step"""
        pass

    @abstractmethod
    def process_step(self) -> ConversionJob:
        """Execute the current conversion step"""
        pass


class OptionalConverter(AbstractConverter, ABC):
    """Base class for optional converters, can be attached to mandatory converters"""

    @abstractmethod
    def check_config(self, config: dict) -> bool:
        """Check if optional converter is activated"""
        pass

    def handle(self, conversion_job: ConversionJob) -> ConversionJob:
        self.conversion_job = conversion_job
        if self.check_config(conversion_job.config):
            self.process_step()
        return self.conversion_job


class ConverterBase(AbstractConverter, ABC):
    """Base class for mandatory converters, can contain optional converters"""

    DESC: str = ""
    PREFIX: str = ""
    POSTFIX: str = ""
    SAVABLE: bool = False

    def __init__(self):
        AbstractConverter.__init__(self)
        self.optional_converters: tuple[list[OptionalConverter], list[OptionalConverter]] = ([], [])
        self.next_step: Optional[ConverterBase] = None
        self.n_step = 1
        self.total_steps = 1

    def handle(self, conversion_job: ConversionJob):
        self.conversion_job = conversion_job
        self.conversion_job.msgqueue.put(QueueMessage.stepmsg(self.n_step, self.total_steps, self.DESC))
        for i in self.optional_converters[0]:
            self.conversion_job = i.handle(self.conversion_job)
        self.conversion_job = self.process_step()
        for i in self.optional_converters[1]:
            self.conversion_job = i.handle(self.conversion_job)
        self.save_to_file()
        if self.next_step is not None:
            self.next_step.handle(self.conversion_job)

    def save_to_file(self):
        """Save state of current conversion to a file (mainly for debugging)"""
        save_intermediate = False
        general = self.conversion_job.config.get("general", False)
        if general is not False:
            save_intermediate = general.get("save_intermediate", False)
        if self.SAVABLE and save_intermediate:
            self.conversion_job.current_step.filename = self.conversion_job.conversion_id + self.PREFIX + self.POSTFIX
            self.conversion_job.current_step.save_file(self.conversion_job.job_assets.sourcefile.directory)

    def chain_append(self, converter: "ConverterBase"):
        """Append a converter to a chain of converters"""
        self.total_steps += 1
        if self.next_step is None:
            converter.total_steps = self.total_steps
            converter.n_step = self.total_steps
            self.next_step = converter
        else:
            self.next_step.chain_append(converter)

    def register_optional(self, converter: OptionalConverter, stage: OptionalStage = OptionalStage.POST):
        """Register an optional converter to execute before or after this converter"""
        if stage == OptionalStage.PRE:
            self.optional_converters[0].append(converter)
        elif stage == OptionalStage.POST:
            self.optional_converters[1].append(converter)



from turandot.gtk.controllers import BaseController, DialogController
from turandot.ui import background, catch_exception
from turandot.gtk import guithread, TypeFilterEnum
from turandot.model import ReferenceSource, ZoteroConnector, ConfigModel
import gi
gi.require_version('Gtk', '3.0')
from gi.repository.Gtk import ComboBoxText, ListStore
from gi.repository import GLib


class ConverterTabController(BaseController):
    """Controller for the ComboBoxes on the 'Converter' tab"""

    def connect_events(self):
        self.view.get_element("reference_source_dropdown").set_change_callback(self.reference_source_callback)
        self.view.get_element("source_file_select_button").connect("clicked", self.select_source_file)
        self.view.get_element("csljson_file_select_button").connect("clicked", self.select_csljson_file)
        self.view.get_element("button_zotero_update_button").connect("clicked", self.update_zotero_libs)

    def select_source_file(self, *args):
        DialogController.db_fed_file_chooser_dialog(
            view=self.view,
            return_entry_id="source_file_entry",
            typefilter=TypeFilterEnum.MARKDOWN
        )

    @guithread
    def reference_source_callback(self, *args):
        entry = self.view.get_element("reference_source_dropdown").get_selected_value()
        if entry.value <= 1:
            self.view.get_element("citation_style_label").hide()
            self.view.get_element("csl_dropdown").hide()
        else:
            self.view.get_element("citation_style_label").show()
            self.view.get_element("csl_dropdown").show()
        if entry == ReferenceSource.ZOTERO:
            self.view.get_element("zotero_lib_label").show()
            self.view.get_element("zotero_lib_dropdown").show()
            self.view.get_element("button_zotero_update_button").show()
            self.update_zotero_libs()
        else:
            self.view.get_element("zotero_lib_label").hide()
            self.view.get_element("zotero_lib_dropdown").hide()
            self.view.get_element("button_zotero_update_button").hide()
        if entry == ReferenceSource.JSON:
            self.view.get_element("csljson_file_label").show()
            self.view.get_element("csljson_file_entry").show()
            self.view.get_element("csljson_file_select_button").show()
        else:
            self.view.get_element("csljson_file_label").hide()
            self.view.get_element("csljson_file_entry").hide()
            self.view.get_element("csljson_file_select_button").hide()

    @background
    @catch_exception
    def update_zotero_libs(self, *args):
        libs = ZoteroConnector(ConfigModel().get_dict()).get_libraries()
        store = ListStore(str, str)
        for i in libs:
            store.append([i["name"], str(i["id"])])
        self.draw_zotero_options(store)

    @guithread
    def draw_zotero_options(self, store: ListStore):
        self.view.get_element("zotero_lib_dropdown").set_model(store)
        self.view.get_element("zotero_lib_dropdown").set_active(0)

    def select_csljson_file(self, *args):
        DialogController.db_fed_file_chooser_dialog(
            view=self.view,
            return_entry_id="csljson_file_entry",
            typefilter=TypeFilterEnum.JSON
        )



from typing import Optional, Callable
from pathlib import Path
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib
from turandot.model import FileSelectPersistApi
from turandot.gtk import TurandotGtkView, FileTypeFilterFactory, TypeFilterEnum
from turandot.gtk.controllers import BaseController


class DialogController(BaseController):
    """Controller to draw various GTK dialogs"""

    def connect_events(self):
        self.view.get_element("confirm_dialog_cancel_button").connect("clicked", self._cancel_confirm)
        self.view.get_element("confirm_dialog").connect("delete_event", self._cancel_confirm)

    def _cancel_confirm(self, *args):
        """Hide delete confirmation dialog on cancel without doing anything"""
        GLib.idle_add(self.view.get_element("confirm_dialog").hide)
        return True

    @staticmethod
    def _file_chooser_dialog(
        view: TurandotGtkView,
        return_entry_id: str,
        init_dir: Optional[Path] = None,
        typefilter: TypeFilterEnum = TypeFilterEnum.ALL,
        title: str = "Choose a file"
    ):
        """Show file chooser dialog"""
        dialog = Gtk.FileChooserDialog(
            title,
            view.window,
            Gtk.FileChooserAction.OPEN,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        )
        FileTypeFilterFactory.apply(typefilter, dialog)
        if init_dir is not None:
            dialog.set_current_folder(str(init_dir))
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            return_entry = view.get_element(return_entry_id)
            filename = Path(dialog.get_filename())
            return_entry.set_text(str(filename))
            FileSelectPersistApi.set(return_entry_id, filename.parent)
        dialog.destroy()
        return True

    @staticmethod
    def db_fed_file_chooser_dialog(
        view: TurandotGtkView,
        return_entry_id: str,
        typefilter: TypeFilterEnum = TypeFilterEnum.ALL,
        title: str = "Choose a file"
    ):
        """Show file chooser dialog, open recently used directory"""
        path = FileSelectPersistApi.get(return_entry_id)
        DialogController._file_chooser_dialog(view, return_entry_id, path, typefilter, title)
        return True

    @staticmethod
    def show_confirm_dialog(view: TurandotGtkView, labeltext: str, callback: Callable):
        """Show delete confirmation dialog"""
        view.get_element("confirm_dialog_ok_button").connect("clicked", callback)
        view.get_element("confirm_dialog_label").set_text(labeltext)
        view.get_element("confirm_dialog").show()



from typing import Optional, Union
from pathlib import Path
import peewee
from bs4 import BeautifulSoup
from turandot.model import DatabaseAsset, ConfigModel
from turandot.model.db.dbmodels import Csl


class CslAsset(DatabaseAsset):
    """Conversion asset representing a CSL file"""

    DB_MODEL = Csl

    def __init__(self, path: Union[Path, str], dbid: Optional[int] = None, expand: bool = False):
        DatabaseAsset.__init__(self, path=path, dbid=dbid, expand=expand)

    def expand(self):
        self._read_path()
        self._read_content()
        self._read_title()

    def _read_title(self):
        try:
            soup = BeautifulSoup(self.content, "xml")
            self.title = str(soup.style.info.title.contents[0])
        except Exception:
            self.title = CslAsset.TITLE_NOT_FOUND_MSG

    def _to_orm(self) -> Csl:
        try:
            entry = Csl.get(dbid=self.dbid)
        except peewee.DoesNotExist:
            return Csl.create(path=self.path)
        entry.path = self.path
        return entry

    @classmethod
    def _from_orm(cls, model: Csl, expand: bool = False):
        return cls(path=model.path, dbid=model.dbid, expand=expand)



from turandot.gtk.controllers.conversionupdater import GtkConversionUpdater
from turandot.gtk.controllers.basecontroller import BaseController
from turandot.gtk.controllers.dialogcontroller import DialogController
from turandot.gtk.controllers.datadropdowncontroller import DataDropdownController
from turandot.gtk.controllers.settingscontroller import SettingsController
from turandot.gtk.controllers.convertertabcontroller import ConverterTabController
from turandot.gtk.controllers.aboutcontroller import AboutController
from turandot.gtk.controllers.exportcontroller import ExportController



from pathlib import Path
from typing import Union
import frontmatter
from turandot.model import TextAsset, ConfigModel


class SourceAsset(TextAsset):
    """Conversion asset representing the Markdown source file"""

    def __init__(self, path: Union[Path, str], expand: bool = False):
        self.metadata = {}
        TextAsset.__init__(self, path=path, expand=expand)

    def _read_content(self):
        enc = ConfigModel().get_key(["general", "encoding"], default="utf8")
        with self.path.open('r', encoding=enc) as f:
            data = frontmatter.load(f)
        self.content = data.content
        self.metadata = data.metadata



from abc import ABC, abstractmethod
from copy import deepcopy

from turandot.ttk.presentations import NotificationReason, DatabaseDropdownObservable
from turandot.ttk.presentations import KvCombobox


class DropdownObserver(ABC):
    """Observer base class to subclass comboboxes from"""

    @abstractmethod
    def update(self, subject: DatabaseDropdownObservable, reason: NotificationReason):
        """Update observer on notification"""
        pass


class DatabaseDropdown(KvCombobox, DropdownObserver):
    """Combobox attached to a database fed observable to change options on notification without new entry option"""

    def __init__(self, master, **kwargs):
        super().__init__(master, **kwargs)

    def update(self, subject: DatabaseDropdownObservable, reason: NotificationReason):
        """Update combobox option on notification"""
        prev_selected = self.get()
        self.clear()
        for i in subject.entries:
            self.add_option(i.dbid, i.title)
        if reason != NotificationReason.INIT and self.check_key(prev_selected):
            self.set(prev_selected)
        else:
            if len(self["values"]) > 0:
                self.current(0)


class DatabaseNewOptDropdown(DatabaseDropdown):
    """Combobox attached to a database fed observable to change options on notification with new entry option"""

    NEW_TEXT = "- create new entry -"

    def update(self, subject: DatabaseDropdownObservable, reason: NotificationReason):
        """Update combobox option on notification"""
        prev_selected = self.get()
        prev_entries = set(self.get_keys())
        self.clear()
        self.add_option(0, DatabaseNewOptDropdown.NEW_TEXT)
        for i in subject.entries:
            self.add_option(i.dbid, i.title)
        if reason == NotificationReason.NEW:
            new_entries = set(self.get_keys())
            diff = new_entries - prev_entries
            self.set(list(diff)[0])
        elif reason in (NotificationReason.CHANGE, NotificationReason.DELETE) and self.check_key(prev_selected):
            self.set(prev_selected)
        else:
            self.current(0)



import os
import multiprocessing
from pathlib import Path
from typing import Optional, List, Union
from dataclasses import dataclass, field
from turandot.model import MessageType


@dataclass
class CopyLog:
    """Log created and copied files to delete afterwards"""

    log: List[tuple] = field(default_factory=list)

    def append(self, elem: tuple):
        """
        Append a copied object to the log
        expects type tuple with `d` as second attribute for a directory or `f` for a file
        """
        self.log.append(elem)

    def delete_all(self):
        """Delete all created files and directories on cleanup"""
        self.log.reverse()
        for i in self.log:
            if i[1] == 'f':
                os.remove(i[0])
            elif i[1] == 'd':
                os.rmdir(i[0])


@dataclass
class ConversionStatus:
    """Container class for all status updates coming from the backend"""
    warnings: list = field(default_factory=list)
    exception: Optional[Exception] = None
    exception_hint: str = "No hint available"
    exception_tb: str = ""
    cause_of_death: MessageType = MessageType.CANCELED


class CompanionData:
    """Data class to attach queue to companion thread during conversion"""

    def __init__(self, queue: multiprocessing.Queue):
        self.queue = queue
        self.copylog = CopyLog()
        self.status = ConversionStatus()



from typing import Optional
from turandot.ui import CatcherStrategy
from turandot.gtk import TurandotGtkView, guithread


class GtkCatcher(CatcherStrategy):

    def __init__(self, view: TurandotGtkView):
        self.view = view
        self._connect_events()

    def _connect_events(self):
        """Connect event listeners to GUI"""
        self.view.elements.get_object("error_dialog").connect("delete_event", self._hide_dialog)
        self.view.elements.get_object("error_dialog_close_button").connect("clicked", self._hide_dialog)

    @guithread
    def handle_exception(self, e: Exception, tb: str):
        self.view.elements.get_object("error_type_dialog_label").set_text(str(type(e).__name__))
        self.view.elements.get_object("error_dialog_text_buffer").set_text(str(e) + "\n\n" + tb)
        self.view.elements.get_object("error_dialog").show()

    @guithread
    def _hide_dialog(self, *args):
        """Hide exception dialog"""
        self.view.elements.get_object("error_dialog").hide()
        # Returning false instead of true avoids a problem where method gets called over and over again
        return False



from turandot.meta import Singleton
from turandot.exceptions import \
    TurandotAssetException, TurandotConversionException, TurandotConnectionException, TurandotCiteprocException
from turandot.sysinfo import SysInfo, OpSys, CpuArch
from turandot.__main__ import launch
__version__ = "3.0.0b2"



from types import SimpleNamespace
from tkinter import *
from tkinter import ttk
from tkinter.font import Font


class TTKStyles:
    """Utility class to attach styles to TTK widgets"""

    def __init__(self):
        lbl = Label(None)  # dummy button from which to extract default font
        self.font_dict = (Font(font=lbl['font'])).actual()  # get settings dict

    def add_styles(self, add_to: Tk):
        self._create_label_styles(add_to)

    def _create_label_styles(self, add_to: Tk):
        h2 = ttk.Style(add_to)
        h2.configure("h2.TLabel", font=(self.font_dict["family"], self.font_dict["size"], "bold"))
        html = ttk.Style(add_to)
        html.configure("html.Text", font=(self.font_dict["family"], self.font_dict["size"], "italic"))

    @staticmethod
    def get_padding() -> SimpleNamespace:
        p = SimpleNamespace()
        p.notebook_frames = "0 5 8 10"
        p.no_title_notebook_frames = "0 15 8 10"
        p.about_notebook_frame = "8 15 8 10"
        p.h2_top = "8 10 0 5"
        p.h2 = "8 20 0 5"
        p.left_labels = {"padx": 8, "pady": 5}
        p.right_entries = {"padx": 0, "pady": 5}
        p.inline_buttons = {"padx": 8, "pady": 20}
        p.right_most_inline_button = {"padx": 0, "pady": 20}
        p.full_process_container = {"padx": 8, "pady": 5}
        return p

    def get_default_font_size(self):
        return self.font_dict["size"]



import uuid
from typing import Optional, Any
from weasyprint import HTML
from pathlib import Path
from turandot import TurandotConversionException, SysInfo, OpSys
from turandot.model import ConverterBase, ConversionJob, QueueMessage


class WeasyprintToPdf(ConverterBase):
    """Use weasyprint to create a PDF file from the HTML step"""

    DESC = "Creating PDF"

    def _write_tmpfile(self) -> Path:
        """Debugging method: Write a temporary file to create a Weasyprint HTML object"""
        enc = self.conversion_job.config.get_key(["general", "encoding"], "utf8")
        tmpfile = self.conversion_job.job_assets.sourcefile.directory / (str(uuid.uuid4()) + ".html")
        with tmpfile.open('w', encoding=enc) as f:
            f.write(self.conversion_job.current_step.content)
        if self.conversion_job.config.get_key(["debug", "weasyprint", "rm_tmp_file"], True):
            self.conversion_job.msgqueue.put(QueueMessage.copymsg((tmpfile, 'f')))
        return tmpfile

    @staticmethod
    def _create_fontconfig() -> Optional["FontConfiguration"]:
        """Create FontConfiguration object to be used by Weasyprint"""
        # sysinfo = SysInfo()
        # if sysinfo.system == OpSys.LINUX:  # FontConfig seems to work fine under Windows
        if True:
            from weasyprint.text.fonts import FontConfiguration
            return FontConfiguration()

    def _create_html_obj(self) -> HTML:
        """Create HTML object for Weasyprint; do it without a temp file if not in debugging mode"""
        filewd = self.conversion_job.job_assets.sourcefile.directory
        enc = self.conversion_job.config.get_key(["general", "encoding"], "utf8")
        if self.conversion_job.config.get_key(["debug", "weasyprint", "use_tmp_file"], False):
            tmpfile = self._write_tmpfile()
            html = HTML(filename=str(tmpfile), encoding=enc)
        else:
            html = HTML(string=self.conversion_job.current_step.content, base_url=str(filewd))
        return html

    def _get_optimizations(self) -> tuple:
        """Read size optimization settings from config file"""
        opt = []
        if self.conversion_job.config.get_key(["processors", "weasyprint_to_pdf", "optimize_size", "images"], True):
            opt.append("images")
        if self.conversion_job.config.get_key(["processors", "weasyprint_to_pdf", "optimize_size", "fonts"], True):
            opt.append("fonts")
        return tuple(opt)

    def _save_pdf(self, html: HTML):
        """Write PDF file"""
        target = self.conversion_job.job_assets.sourcefile.path.with_suffix(".pdf")
        opt = self._get_optimizations()
        fc = WeasyprintToPdf._create_fontconfig()
        try:
            doc = html.render(optimize_size=opt, font_config=fc)
        except IndexError:
            raise TurandotConversionException("Index Error in fontconfig, probably a syntax error in a @font-face rule")
        doc.metadata.generator = "Turandot"
        doc.write_pdf(target=target)

    def process_step(self) -> ConversionJob:
        html = self._create_html_obj()
        self._save_pdf(html)
        return self.conversion_job



from tkinter import messagebox
from turandot.ttk.controllers import ControllerBase


class ExceptionController(ControllerBase):
    """Controller to draw an exception onto the frontend"""

    def draw_exception(self, e: Exception, tb: str):
        """Draw exception onto the frontend"""
        exceptionstring = str(e)
        errortype = str(type(e).__name__)
        messagebox.showerror(title="Error occurred", message=exceptionstring)
        self.view.widgets["error_text_view"].delete(1.0, "end")
        self.view.widgets["error_text_view"].insert(1.0, f"{exceptionstring}\n\n{tb}")
        self.view.widgets["error_type_label_value"].set(errortype)
        self.view.widgets["error_text_view"].grid()
        self.view.widgets["error_type_desc_label"].grid()
        self.view.widgets["error_type_label"].grid()
        self.view.widgets["error_title"].grid()



from tkinter import *
from tkinter import ttk

from turandot.ttk.presentations import KvCombobox
from turandot.ttk.view.styles import TTKStyles
from turandot.ttk.view import ViewComponent


class ConverterTab(ViewComponent):
    """View component for the 'Convert' tab"""

    def _create_source_widgets(self, cframe: ttk.Frame):
        # Source file title
        test = ttk.Label(cframe, text="Source", style="h2.TLabel", padding=TTKStyles.get_padding().h2_top)
        test.grid(row=0, columnspan=2, sticky=(W, E))
        # Source file select entry
        ttk.Label(cframe, text="Source file:")\
            .grid(row=1, column=0, sticky=W, **TTKStyles.get_padding().left_labels)
        self.widgets["converter_source_file_entry_value"] = StringVar(value="")
        self.widgets["converter_source_file_entry"] = ttk.Entry(
            cframe, textvariable=self.widgets["converter_source_file_entry_value"]
        )
        self.widgets["converter_source_file_entry"].grid(
            row=1, column=1, sticky=(W, E), **TTKStyles.get_padding().right_entries
        )
        # Source file select button
        self.widgets["converter_source_pick_button"] = ttk.Button(cframe, text="Select file")
        self.widgets["converter_source_pick_button"].grid(
            row=2, columnspan=2, sticky=E, **TTKStyles.get_padding().right_entries
        )
        # Template dropdown
        ttk.Label(cframe, text="Template:").grid(row=3, column=0, sticky=W, **TTKStyles.get_padding().left_labels)
        self.widgets["converter_template_dropdown_frame"] = ttk.Frame(cframe)
        self.widgets["converter_template_dropdown_frame"].grid(row=3, column=1, sticky=(W, E))
        self.widgets["converter_template_dropdown_frame"].columnconfigure(0, weight=1)
        # Algorithm dropdown
        ttk.Label(cframe, text="Algorithm:").grid(row=4, column=0, sticky=W, **TTKStyles.get_padding().left_labels)
        self.widgets["converter_algorithm_dropdown_frame"] = ttk.Frame(cframe)
        self.widgets["converter_algorithm_dropdown_frame"].grid(row=4, column=1, sticky=(W, E))
        self.widgets["converter_algorithm_dropdown_frame"].columnconfigure(0, weight=1)

    def _create_reference_widgets(self, cframe: ttk.Frame):
        # Reference title
        ttk.Label(cframe, text="References", style="h2.TLabel", padding=TTKStyles.get_padding().h2)\
            .grid(row=5, columnspan=2, sticky=(W, E))
        # Reference source dropdown
        ttk.Label(cframe, text="Reference data source:").grid(
            row=6, column=0, sticky=W, **TTKStyles.get_padding().left_labels
        )
        self.widgets["converter_reference_source_dropdown_frame"] = ttk.Frame(cframe)
        self.widgets["converter_reference_source_dropdown_frame"].grid(row=6, column=1, sticky=(W, E))
        self.widgets["converter_reference_source_dropdown_frame"].columnconfigure(0, weight=1)
        # CSL dropdown
        self.widgets["converter_csl_label"] = ttk.Label(cframe, text="Citation style:")
        self.widgets["converter_csl_label"].grid(row=7, column=0, sticky=W, **TTKStyles.get_padding().left_labels)
        self.widgets["converter_csl_dropdown_frame"] = ttk.Frame(cframe)
        self.widgets["converter_csl_dropdown_frame"].grid(row=7, column=1, sticky=(W, E))
        self.widgets["converter_csl_dropdown_frame"].columnconfigure(0, weight=1)
        # Zotero widgets
        self.widgets["converter_zotero_lib_label"] = ttk.Label(cframe, text="Zotero library:")
        self.widgets["converter_zotero_lib_label"].grid(
            row=8, column=0, sticky=W, **TTKStyles.get_padding().left_labels
        )
        self.widgets["converter_zotero_lib_dropdown_frame"] = ttk.Frame(cframe)
        self.widgets["converter_zotero_lib_dropdown_frame"].grid(row=8, column=1, sticky=(W, E))
        self.widgets["converter_zotero_lib_dropdown_frame"].columnconfigure(0, weight=1)
        self.widgets["converter_zotero_lib_dropdown"] = KvCombobox(
            self.widgets["converter_zotero_lib_dropdown_frame"], state="readonly"
        )
        self.widgets["converter_zotero_lib_dropdown"].grid(
            row=0, column=0, sticky=(W, E), **TTKStyles.get_padding().right_entries
        )
        self.widgets["converter_zotero_update_button"] = ttk.Button(cframe, text="Update libraries")
        self.widgets["converter_zotero_update_button"].grid(
            row=9, columnspan=2, sticky=E, **TTKStyles.get_padding().right_entries
        )
        # CSLJSON widgets
        self.widgets["converter_csljson_label"] = ttk.Label(cframe, text="CSLJSON file:")
        self.widgets["converter_csljson_label"].grid(row=10, column=0, sticky=W, **TTKStyles.get_padding().left_labels)
        self.widgets["converter_csljson_entry_value"] = StringVar(value="")
        self.widgets["converter_csljson_entry"] = ttk.Entry(
            cframe, textvariable=self.widgets["converter_csljson_entry_value"]
        )
        self.widgets["converter_csljson_entry"].grid(row=10, column=1, sticky=(W, E))
        self.widgets["converter_csljson_pick_button"] = ttk.Button(cframe, text="Select file")
        self.widgets["converter_csljson_pick_button"].grid(
            row=11, columnspan=2, sticky=E, **TTKStyles.get_padding().right_entries
        )

    def _create_export_buttons(self, cframe: ttk.Frame):
        container = ttk.Frame(cframe)
        container.grid(row=12, columnspan=2, sticky=E)
        self.widgets["converter_cancel_export_button"] = ttk.Button(container, text="Cancel export")
        self.widgets["converter_cancel_export_button"].grid(
            row=0, column=0, sticky=E, **TTKStyles.get_padding().inline_buttons
        )
        self.widgets["converter_export_doc_button"] = ttk.Button(container, text="Export document")
        self.widgets["converter_export_doc_button"].grid(
            row=0, column=1, sticky=E, **TTKStyles.get_padding().right_most_inline_button
        )

    def create(self, add_to: ttk.Frame) -> ttk.Frame:
        cframe = ttk.Frame(add_to)
        cframe.grid(column=0, row=0, sticky=(N, S, W, E))
        cframe.columnconfigure(0, weight=0)
        cframe.columnconfigure(1, weight=1)
        self._create_source_widgets(cframe)
        self._create_reference_widgets(cframe)
        self._create_export_buttons(cframe)
        return cframe



from turandot.gtk.presentations.dbdropdown import DatabaseDropdown
from turandot.gtk.presentations.enumdropdown import EnumDropdown
from turandot.gtk.presentations.configpresentation import SettingsControlBase, TitleControl, TextControl, SwitchControl, SpinControl, TextEnumControl



from tkinter import *

from turandot.model import ReferenceSource, ConversionAlgorithm, ZoteroConnector, ConfigModel
from turandot.ui import EnumTranslations, background, catch_exception
from turandot.ttk.controllers import ControllerBase
from turandot.ttk.filetypes import FileTypes
from turandot.ttk.presentations import EnumCombobox, DbFedFilePicker
from turandot.ttk.view import TTKStyles
from turandot.ttk.view import TurandotTtkView


class ConverterController(ControllerBase):
    """Controller to add and handle comboboxes on controller tab"""

    def __init__(self, view: TurandotTtkView):
        super().__init__(view)
        self._attach_button_events()
        self._add_enum_dropdowns()

    def _attach_button_events(self):
        """Handle button press events"""
        self.view.widgets["converter_zotero_update_button"].bind("<Button-1>", self._update_zotero_libs)
        self.view.widgets["converter_cancel_export_button"].state(["disabled"])
        self.view.widgets["converter_source_pick_button"].bind("<Button-1>", self._source_picker_callback)
        self.view.widgets["converter_csljson_pick_button"].bind("<Button-1>", self._csljson_picker_callback)

    @background
    def _source_picker_callback(self, *args):
        """Draw file picker dialog on select source button"""
        DbFedFilePicker.draw(self.view, "converter_source_file_entry_value", FileTypes.markdown)
        return "break"

    @background
    def _csljson_picker_callback(self, *args):
        """Draw file picker dialog on select csljson button"""
        DbFedFilePicker.draw(self.view, "converter_csljson_entry_value", FileTypes.json)
        return "break"

    def _add_enum_dropdowns(self):
        """Add various comboboxes from enums"""
        m = self.view.widgets["converter_reference_source_dropdown_frame"]
        self.view.widgets["converter_reference_source_dropdown"] = EnumCombobox(
            m, ReferenceSource, EnumTranslations.textentries, state="readonly"
        )
        self.view.widgets["converter_reference_source_dropdown"].grid(
            row=0, column=0, sticky=(W, E), **TTKStyles.get_padding().right_entries
        )
        self.view.widgets["converter_reference_source_dropdown"].set_callback(self._ref_source_callback)
        self.view.widgets["converter_reference_source_dropdown"].current(0)
        m = self.view.widgets["converter_algorithm_dropdown_frame"]
        self.view.widgets["converter_algorithm_dropdown"] = EnumCombobox(
            m, ConversionAlgorithm, EnumTranslations.textentries, state="readonly"
        )
        self.view.widgets["converter_algorithm_dropdown"].grid(
            row=0, column=0, sticky=(W, E), **TTKStyles.get_padding().right_entries
        )
        self.view.widgets["converter_algorithm_dropdown"].current(0)

    def _ref_source_callback(self, *args):
        """Draw various widgets specific to the selected reference source"""
        refsource = self.view.widgets["converter_reference_source_dropdown"].get()
        if refsource.value <= 1:
            self.view.widgets["converter_csl_label"].grid_remove()
            self.view.widgets["converter_csl_dropdown_frame"].grid_remove()
        else:
            self.view.widgets["converter_csl_label"].grid()
            self.view.widgets["converter_csl_dropdown_frame"].grid()
        if refsource == ReferenceSource.ZOTERO:
            self.view.widgets["converter_zotero_lib_dropdown_frame"].grid()
            self.view.widgets["converter_zotero_lib_label"].grid()
            self.view.widgets["converter_zotero_update_button"].grid()
            self._update_zotero_libs()
        else:
            self.view.widgets["converter_zotero_lib_dropdown_frame"].grid_remove()
            self.view.widgets["converter_zotero_lib_label"].grid_remove()
            self.view.widgets["converter_zotero_update_button"].grid_remove()
        if refsource == ReferenceSource.JSON:
            self.view.widgets["converter_csljson_label"].grid()
            self.view.widgets["converter_csljson_entry"].grid()
            self.view.widgets["converter_csljson_pick_button"].grid()
        else:
            self.view.widgets["converter_csljson_label"].grid_remove()
            self.view.widgets["converter_csljson_entry"].grid_remove()
            self.view.widgets["converter_csljson_pick_button"].grid_remove()

    @background
    @catch_exception
    def _update_zotero_libs(self, *args):
        """Update comboboxes based on available zotero libraries"""
        self.view.widgets["converter_zotero_lib_dropdown"].clear()
        libs = ZoteroConnector(ConfigModel().get_dict()).get_libraries()
        for i in libs:
            self.view.widgets["converter_zotero_lib_dropdown"].add_option(i["id"], i["name"])
        self.view.widgets["converter_zotero_lib_dropdown"].current(0)
        return "break"



from turandot.gtk.gtkenums import DropdownPreSelections
from turandot.gtk.decorators import guithread
from turandot.gtk.filetypefilters import TypeFilterEnum, FileTypeFilterFactory
from turandot.gtk.view import TurandotGtkView
from turandot.gtk.gtkcatcher import GtkCatcher
from turandot.gtk.__main__ import TurandotGtk



from md_citeproc import NotationStyle, OutputStyle

from turandot.model import ModelUtils
from turandot.ui import FrontendUtils
from turandot.ttk.controllers import ControllerBase
from turandot.ttk.presentations import TitleControl, TextControl, SpinControl, SwitchControl, TextEnumControl
from turandot.ttk.view import TurandotTtkView


class SettingsController(ControllerBase):
    """Draw settings presentations to settings tab"""

    def __init__(self, view: TurandotTtkView):
        super().__init__(view)
        self._fill_config_file_control()
        self._create_settings_controls()

    def _fill_config_file_control(self):
        """Draw settings file path to GUI"""
        self.view.widgets["config_file_location_entry_value"].set(ModelUtils.get_config_file())
        self.view.widgets["config_dir_open_button"].bind("<Button-1>", SettingsController._open_config_dir)

    @staticmethod
    def _open_config_dir(*args):
        """Open system file manager on config folder"""
        FrontendUtils.fm_open_path(ModelUtils.get_config_dir())
        return "break"

    def _create_settings_controls(self):
        """Draw settings presentations onto GUI"""
        add_to = self.view.widgets.get("settings_container")
        TitleControl("General", add_to, 3)
        SwitchControl("Remember file input path:", ["general", "file_select_persistence"], add_to, 4)
        SwitchControl("Save intermediate files:", ["general", "save_intermediate"], add_to, 5)
        TitleControl("Zotero", add_to, 6)
        SpinControl("BetterBibtex Port:", ['api', 'zotero', 'port'], add_to, 7)
        TitleControl("Table of contents", add_to, 8)
        TextControl("TOC Marker:", ['processors', 'convert_to_html', 'markdown_ext', 'toc', 'marker'], add_to, 9)
        TitleControl("Citeproc", add_to, 10)
        TextControl("Locale:", ['processors', 'convert_to_html', 'markdown_ext', 'md_citeproc', 'locale'], add_to, 11),
        TextEnumControl(
            "Notation Style:",
            ['processors', 'convert_to_html', 'markdown_ext', 'md_citeproc', 'notation'],
            add_to=add_to,
            row=12,
            model=NotationStyle
        )
        TextEnumControl(
            "Output Style:",
            ['processors', 'convert_to_html', 'markdown_ext', 'md_citeproc', 'output'],
            add_to=add_to,
            row=13,
            model=OutputStyle
        )
        TextControl("Footnotes Marker:",
                    ['processors', 'convert_to_html', 'markdown_ext', 'md_citeproc', 'footnotes_token'], add_to, 14)
        TextControl("Bibliography Marker:",
                    ['processors', 'convert_to_html', 'markdown_ext', 'md_citeproc', 'bibliography_token'], add_to, 15)
        TitleControl("Optional processors", add_to, 16)
        SwitchControl("Unified math markers:", ['opt_processors', 'unified_math_block_marker', 'enable'], add_to, 17)
        SwitchControl("TOC pagination containers:",
                      ['opt_processors', 'toc_pagination_containers', 'enable'], add_to, 18)



from tkinter import *
from tkinter import ttk

from turandot.ttk.view.styles import TTKStyles
from turandot.ttk.view import ViewComponent


class TemplateTab(ViewComponent):
    """Create view component for 'Template' tab"""

    def _create_file_template_selector(self, tframe: ttk.Frame):
        ttk.Label(tframe, text="Entry to edit:").grid(
            row=0, column=0, sticky=W, **TTKStyles.get_padding().left_labels
        )
        self.widgets["template_editor_dropdown_frame"] = ttk.Frame(tframe)
        self.widgets["template_editor_dropdown_frame"].grid(row=0, column=1, sticky=(W, E))
        self.widgets["template_editor_dropdown_frame"].columnconfigure(0, weight=1)
        ttk.Label(tframe, text="Template file:").grid(
            row=1, column=0, sticky=W, **TTKStyles.get_padding().left_labels
        )
        self.widgets["tmpl_base_file_enty_value"] = StringVar(value="")
        self.widgets["tmpl_base_file_enty"] = ttk.Entry(tframe, textvariable=self.widgets["tmpl_base_file_enty_value"])
        self.widgets["tmpl_base_file_enty"].grid(
            row=1, column=1, sticky=(W, E), **TTKStyles.get_padding().right_entries
        )
        self.widgets["tmpl_base_select_button"] = ttk.Button(tframe, text="Select file")
        self.widgets["tmpl_base_select_button"].grid(
            row=2, columnspan=2, sticky=E, **TTKStyles.get_padding().right_entries
        )

    def _create_template_switches(self, tframe: ttk.Frame):
        switchframe = ttk.Frame(tframe)
        switchframe.grid(row=3, columnspan=2, sticky=(W, E))
        switchframe.columnconfigure(0, weight=1)
        switchframe.columnconfigure(1, weight=0)
        ttk.Label(switchframe, text="Allow Jinja Templating:").grid(
            row=0, column=0, sticky=W, **TTKStyles.get_padding().left_labels
        )
        self.widgets["tmpl_allow_jinja_value"] = BooleanVar(value=False)
        self.widgets["tmpl_allow_jinja"] = ttk.Checkbutton(
            switchframe, text="", variable=self.widgets["tmpl_allow_jinja_value"]
        )
        self.widgets["tmpl_allow_jinja"].grid(row=0, column=1, sticky=E)
        ttk.Label(switchframe, text="Allow Mako Templating (DANGEROUS!):").grid(
            row=1, column=0, sticky=W, **TTKStyles.get_padding().left_labels
        )
        self.widgets["tmpl_allow_mako_value"] = BooleanVar(value=False)
        self.widgets["tmpl_allow_mako"] = ttk.Checkbutton(
            switchframe, text="", variable=self.widgets["tmpl_allow_mako_value"]
        )
        self.widgets["tmpl_allow_mako"].grid(row=1, column=1, sticky=E)

    def _create_save_del_buttons(self, tframe: ttk.Frame):
        buttonframe = ttk.Frame(tframe)
        buttonframe.grid(row=4, columnspan=2, sticky=E)
        buttonframe.columnconfigure(0, weight=1)
        self.widgets["tmpl_delete_button"] = ttk.Button(buttonframe, text="Delete entry")
        self.widgets["tmpl_delete_button"].state(["disabled"])
        self.widgets["tmpl_delete_button"].grid(
            row=0, column=0, sticky=E, **TTKStyles.get_padding().inline_buttons
        )
        self.widgets["tmpl_save_button"] = ttk.Button(buttonframe, text="Save entry")
        self.widgets["tmpl_save_button"].grid(
            row=0, column=1, sticky=E, **TTKStyles.get_padding().right_most_inline_button
        )

    def create(self, add_to: ttk.Frame) -> ttk.Frame:
        tframe = ttk.Frame(add_to)
        tframe.grid(column=0, row=0, sticky=N+S+E+W)
        tframe.columnconfigure(0, weight=0)
        tframe.columnconfigure(1, weight=1)
        self._create_file_template_selector(tframe)
        self._create_template_switches(tframe)
        self._create_save_del_buttons(tframe)
        return tframe
